#
# COPYRIGHT
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
import imageio
import mrcfile
import numpy as np
from PIL import Image


def image_read(image_path, region=None):
    image_path = str(image_path)
    if image_path.endswith(("jpg", "png")):
        if not is_single_channel(image_path):
            raise Exception("Not supported image format. Movie files are not supported")
            return None
        img = imageio.imread(image_path, pilmode="L", as_gray=True)
        img = img.astype(np.uint8)
    elif image_path.endswith(("tiff", "tif")):
        img = imageio.imread(image_path)
    # img = np.flipud(img)
    elif image_path.endswith(("mrc", "mrcs")):
        if not is_single_channel(image_path):
            raise Exception("Not supported image format. Movie files are not supported")
            return None
        img = read_mrc(image_path)
    else:
        raise Exception("Not supported image format: " + image_path)

    # OpenCV has problems with some datatypes
    if np.issubdtype(img.dtype, np.uint32):
        img = img.astype(dtype=np.float64)

    if np.issubdtype(img.dtype, np.float16):
        img = img.astype(dtype=np.float32)

    if np.issubdtype(img.dtype, np.uint16):
        img = img.astype(dtype=np.float32)

    if region is not None:
        return img[region[1], region[0]]
    return img


def image_write(image_path, image):
    if image_path.endswith(("jpg", "png")):
        imageio.imwrite(image_path, image)
    elif image_path.endswith(("tiff", "tif")):
        image = np.float32(image)
        imageio.imwrite(image_path, image)
    elif image_path.endswith(("mrc","mrcs")):
        image = np.flipud(image)
        with mrcfile.new(image_path, overwrite=True) as mrc:
            mrc.set_data(np.float32(image))


def is_single_channel(image_path):
    if image_path.endswith(("jpg", "png", "tiff", "tif")):
        im = Image.open(image_path)
        if len(im.size) > 2:
            return False

    if image_path.endswith(("mrc","mrcs")):
        with mrcfile.mmap(image_path, permissive=True, mode="r") as mrc:
            if mrc.header.nz > 1:
                return False

    return True


def read_width_height(image_path):
    if image_path.endswith(("jpg", "png", "tiff", "tif")):
        im = Image.open(image_path)
        width, height = im.size
        return width, height

    if image_path.endswith(("mrc","mrcs")):
        with mrcfile.mmap(image_path, permissive=True, mode="r") as mrc:
            img_height = mrc.header.ny
            img_width = mrc.header.nx
        return img_width, img_height


def read_mrc(image_path):
    with mrcfile.mmap(image_path, permissive=True, mode="r") as mrc:
        mrc_image_data = mrc.data
        mrc_image_data = np.squeeze(mrc_image_data)
        mrc_image_data = np.flipud(mrc_image_data)

    return mrc_image_data


def get_tile_coordinates(imgw, imgh, num_patches, patchxy, overlap=0):
    patch_width = int(imgw / num_patches)
    patch_height = int(imgh / num_patches)
    region_from_x = int(patchxy[0] * patch_width)
    region_to_x = int((patchxy[0] + 1) * patch_width)
    region_from_y = int(patchxy[1] * patch_height)
    region_to_y = int((patchxy[1] + 1) * patch_height)
    overlap = int(overlap)
    if patchxy[0] == 0:
        region_to_x = region_to_x + 2 * overlap
    elif patchxy[0] == (num_patches - 1):
        region_from_x = region_from_x - 2 * overlap
    else:
        region_from_x = region_from_x - overlap
        region_to_x = region_to_x + overlap

    if patchxy[1] == 0:
        region_to_y = region_to_y + 2 * overlap
    elif patchxy[1] == (num_patches - 1):
        region_from_y = region_from_y - 2 * overlap
    else:
        region_from_y = region_from_y - overlap
        region_to_y = region_to_y + overlap

    if region_to_x > imgw:
        region_to_x = imgw
    if region_to_y > imgh:
        region_to_y = imgh

    tile = np.s_[region_from_x:region_to_x, region_from_y:region_to_y]

    return tile



#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

import argparse
import datetime
import difflib
import json
import multiprocessing
import os

import numpy as np
import terminaltables
from keras.utils import multi_gpu_model

from cryolo import predict
from . import config_tools
from . import imagereader
from .frontend import YOLO
from gooey import Gooey, GooeyParser


"""
This scripts evaluates a model based on ground truth box files.

Per annotation folder it calculates:
    As table:
    - Optimal threshold based on F1 score
    - AUC 
    - Precision using optimal threshold, 0.3, 0.2 and 0.1
    - Recall using optimal threshold, 0.3, 0.2 and 0.1
    - F1 score using optimal threshold, 0.3, 0.2 and 0.1

Morover it calculates for the full data:
 
    - Optimal threshold based on F1 score
    - AUC 
    As table 2:
    - Precision using optimal threshold, 0.3, 0.2 and 0.1
    - Recall using optimal threshold, 0.3, 0.2 and 0.1
    - F1 score using optimal threshold, 0.3, 0.2 and 0.1
    
    

    
"""


argparser = argparse.ArgumentParser(
    description="Calculate common statistics using box files"
)

def create_parser(parser):
    required_group = parser.add_argument_group(
        "Required arguments", "These options are required to run crYOLO evaluation"
    )
    required_group.add_argument("-c",
                                "--config",
                                required=True,
                                help="Path to configuration file (.json)",
                                widget="FileChooser",
                                gooey_options={
                                    "validator": {
                                        "test": 'user_input.endswith("json")',
                                        "message": "File has to end with .json!",
                                    },
                                    "wildcard": "*.json"
                                },
                                )

    required_group.add_argument("-w",
                                "--weights",
                                required=True,
                                help="Path to trained model (.h5 file)",
                                widget="FileChooser",
                                gooey_options={
                                    "validator": {
                                        "test": 'user_input.endswith("h5")',
                                        "message": "File has to end with .h5!",
                                    },
                                    "wildcard": "*.h5"
                                },
                                )

    required_group.add_argument("-r",
                                "--runfile",
                                help="Evaluate you model based on the runfile (.json). "
                                                        "You find your runfile connected to your model "
                                                        "in the runfiles folder where the training command was started. "
                                                        "It contains a list of the validation images "
                                                        "and will use them to evaluate your model.",
                                widget="FileChooser",
                                gooey_options={
                                    "wildcard": "*.json"
                                }
                                )

    required_group.add_argument("-o",
                                "--output",
                                help="Select path where the results are written as .html ",
                                default="result_evaluation.html",
                                widget="FileSaver",
                                gooey_options={
                                    "validator": {
                                        "test": 'user_input.endswith("html")',
                                        "message": "File has to end with .html!",
                                    },
                                    "default_file": "result_evaluation.html"
                                },
                                )

    optional_group = parser.add_argument_group(
        "Optional arguments", "These options are optional to run crYOLO evaluation"
    )

    optional_group.add_argument(
        "-i", "--images", help="path folder with test images"
    )

    optional_group.add_argument(
        "-b", "--boxfiles", help="path to folder with box files (ground truth)"
    )

    optional_group.add_argument(
        "-g", "--gpu", default=0, type=int, help="Specifiy which gpu should be used."
    )

def get_parser():
    parser = GooeyParser(
        description="Evlaute your trained models",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    create_parser(parser)
    return parser


def _main_():
    import sys

    # if sys.argv[1] == "gui":
    if len(sys.argv) >= 2:
        if not "--ignore-gooey" in sys.argv:
            sys.argv.append("--ignore-gooey")

    # r'^\d+ particles are found in .* \( (\d+) % \)$'
    kwargs = {"terminal_font_family": "monospace",
              "richtext_controls": True}
    Gooey(
        main,
        program_name="crYOLO evaluation",
        image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        **kwargs
    )()

def main(args=None):

    if args is None:
        parser = get_parser()
        args = parser.parse_args()

    path_config = args.config
    path_weights = args.weights
    path_output = args.output
    path_runfile = None
    if args.runfile:
        path_runfile = args.runfile
        path_boxfiles = None
        path_images = None
    else:
        if args.boxfiles is not None and args.images is not None:
            path_boxfiles = args.boxfiles
            path_images = args.images
        else:
            import sys

            sys.exit(
                "Please specify either the path to your validation data with -b (--boxfiles) and -i (--images)"
                " or the path to the runfile -r (--runfile)"
            )

    os.environ["CUDA_VISIBLE_DEVICES"] = str(args.gpu)

    # Create evaluation datasets
    datasets = prepare_datasets(
        path_config, path_weights, path_boxfiles, path_runfile, path_images
    )

    # Calculate statistics
    calculate_all_statistics(datasets)

    # Calculate optimal F1 F2 threshold
    arange = np.arange(start=0, stop=1.001, step=0.001)
    f1_array = np.zeros(len(arange))
    f1_N_array = np.zeros(len(arange))
    f2_array = np.zeros(len(arange))
    f2_N_array = np.zeros(len(arange))
    for foldername, dataset in datasets.items():
        if "results" in dataset:
            result_dict = dataset["results"]
            for i, thr in enumerate(arange):
                if str(thr) in result_dict:
                    f1_array[i] += result_dict[str(thr)][0]
                    f1_N_array[i] += 1
                    f2_array[i] += result_dict[str(thr)][5]
                    f2_N_array[i] += 1
    for i in range(len(f1_array)):
        if f1_N_array[i] > 0:
            f1_array[i] = f1_array[i] / f1_N_array[i]
        if f2_N_array[i] > 0:
            f2_array[i] = f2_array[i] / f2_N_array[i]
    best_f1_thresh_index = np.argmax(f1_array)
    best_f1_thresh = result_dict["THRESH_LIST"][best_f1_thresh_index]

    best_f2_thresh_index = np.argmax(f2_array)
    best_f2_thresh = result_dict["THRESH_LIST"][best_f2_thresh_index]

    # Print statistics
    table_data = []
    template_column_folder = "\033[1;31m{0}\033[0m"
    template_column_value = "\033[1;34m{0}\033[0m"
    template_row_value = "\033[100m{0}\033[0m"
    template_row_white = "\033[0m{0}\033[0m"
    template = template_row_value
    headings = [
        "AUC",
        "Topt",
        "R\nTopt",
        "R\n0.3",
        "R\n0.2",
        # "R\n0.1",
        "P\nTopt",
        "P\n0.3",
        "P\n0.2",
        # "P\n0.1",
        "F1\nTopt",
        "F1\n0.3",
        "F1\n0.2",
        # "F1\n0.1"
        "IOU\nTopt",
        "IOU\n0.3",
        "IOU\n0.2",
    ]
    headings = [
        "\n".join(
            [template_column_value.format(entry2) for entry2 in entry.split("\n")]
        )
        if index % 2 == 1
        else entry
        for index, entry in enumerate(headings)
    ]
    headings.insert(0, template_column_folder.format("Folder"))
    table_data.append(headings)
    n = 0
    auc_avg = 0
    topt_avg = 0
    pTopt_avg = 0
    p03_avg = 0
    p02_avg = 0
    p01_avg = 0
    rTopt_avg = 0
    r03_avg = 0
    r02_avg = 0
    r01_avg = 0
    fTopt_avg = 0
    f03_avg = 0
    f02_avg = 0
    f01_avg = 0
    iouTopt_avg = 0
    iou03_avg = 0
    iou02_avg = 0

    for foldername, dataset in datasets.items():
        result_dict = dataset["results"]

        AUC = "{0:.2f}".format(result_dict["AUC"])
        auc_avg += result_dict["AUC"]

        topt = "{0:.2f}".format(result_dict["MAX_F1_THRESH"])
        topt_avg += result_dict["MAX_F1_THRESH"]

        pTopt = "{0:.2f}".format(result_dict[str(result_dict["MAX_F1_THRESH"])][1])
        pTopt_avg += result_dict[str(result_dict["MAX_F1_THRESH"])][1]

        p03 = "{0:.2f}".format(result_dict[str(0.3)][1])
        p03_avg += result_dict[str(0.3)][1]

        p02 = "{0:.2f}".format(result_dict[str(0.2)][1])
        p02_avg += result_dict[str(0.2)][1]

        p01 = "{0:.2f}".format(result_dict[str(0.1)][1])
        p01_avg += result_dict[str(0.1)][1]

        rTopt = "{0:.2f}".format(result_dict[str(result_dict["MAX_F1_THRESH"])][2])
        rTopt_avg += result_dict[str(result_dict["MAX_F1_THRESH"])][2]

        r03 = "{0:.2f}".format(result_dict[str(0.3)][2])
        r03_avg += result_dict[str(0.3)][2]

        r02 = "{0:.2f}".format(result_dict[str(0.2)][2])
        r02_avg += result_dict[str(0.2)][2]

        r01 = "{0:.2f}".format(result_dict[str(0.1)][2])
        r01_avg += result_dict[str(0.1)][2]

        fTopt = "{0:.2f}".format(result_dict[str(result_dict["MAX_F1_THRESH"])][0])
        fTopt_avg += result_dict[str(result_dict["MAX_F1_THRESH"])][0]

        f03 = "{0:.2f}".format(result_dict[str(0.3)][0])
        f03_avg += result_dict[str(0.3)][0]

        f02 = "{0:.2f}".format(result_dict[str(0.2)][0])
        f02_avg += result_dict[str(0.2)][0]

        f01 = "{0:.2f}".format(result_dict[str(0.1)][0])
        f01_avg += result_dict[str(0.1)][0]

        iouTopt = "{0:.2f}".format(result_dict[str(result_dict["MAX_F1_THRESH"])][3])
        iouTopt_avg += result_dict[str(result_dict["MAX_F1_THRESH"])][3]

        iou03 = "{0:.2f}".format(result_dict[str(0.3)][3])
        iou03_avg += result_dict[str(0.3)][3]

        iou02 = "{0:.2f}".format(result_dict[str(0.2)][3])
        iou02_avg += result_dict[str(0.2)][3]

        table_row = [
            AUC,
            topt,
            rTopt,
            r03,
            r02,
            # r01,
            pTopt,
            p03,
            p02,
            # p01,
            fTopt,
            f03,
            f02,
            # f01
            iouTopt,
            iou03,
            iou02,
        ]

        table_row = [
            "\n".join(
                [
                    template.format(template_column_value.format(entry2))
                    for entry2 in entry.split("\n")
                ]
            )
            if index % 2 == 1
            else template.format(entry)
            for index, entry in enumerate(table_row)
        ]
        table_row.insert(0, template.format(template_column_folder.format(foldername)))

        table_data.append(table_row)

        if template == template_row_value:
            template = template_row_white
        else:
            template = template_row_value
        n = n + 1
    table_row = [
        "{0:.2f}".format(auc_avg / n),
        "{0:.2f}".format(topt_avg / n),
        "{0:.2f}".format(rTopt_avg / n),
        "{0:.2f}".format(r03_avg / n),
        "{0:.2f}".format(r02_avg / n),
        # '{0:.2f}'.format(r01_avg/n),
        "{0:.2f}".format(pTopt_avg / n),
        "{0:.2f}".format(p03_avg / n),
        "{0:.2f}".format(p02_avg / n),
        # '{0:.2f}'.format(p01_avg/n),
        "{0:.2f}".format(fTopt_avg / n),
        "{0:.2f}".format(f03_avg / n),
        "{0:.2f}".format(f02_avg / n),
        # '{0:.2f}'.format(f01_avg/n)
        "{0:.2f}".format(iouTopt_avg / n),
        "{0:.2f}".format(iou03_avg / n),
        "{0:.2f}".format(iou02_avg / n),
    ]
    table_row = [
        "\n".join(
            [
                template.format(template_column_value.format(entry2))
                for entry2 in entry.split("\n")
            ]
        )
        if index % 2 == 1
        else template.format(entry)
        for index, entry in enumerate(table_row)
    ]
    table_row.insert(0, template.format(template_column_folder.format("AVERAGE")))

    table_data.append(table_row)

    table = terminaltables.AsciiTable(table_data)
    import hashlib

    hasher = hashlib.md5()
    with open(path_weights, "rb") as afile:
        buf = afile.read()
        hasher.update(buf)
    md5sum = hasher.hexdigest()

    if args.runfile:
        config_str = "Config: " + path_config + " Runfile: " + path_runfile
    else:
        config_str = "Config: " + path_config + " Valid images:" + path_images + " Valid annotations: " + path_boxfiles

    out_res = "##########################"+\
          "\n"+\
          "### Summary evaluation ###"+\
          "\n"+\
          "##########################"+\
          "\n"+\
          "Date: "+ str(datetime.date.today())+\
          "\n"+\
          "Model: "+ path_weights+ "Model checksum (MD5):"+ str(md5sum)+\
          "\n"+\
          config_str+\
          "\n"+\
          "R = Recall. Higher is better. Range 0 - 1"+\
          "\n"+\
          "P = Precision. Higher is better. Range 0 -1"+\
          "\n"+\
          "F1 = F1 Score, harmonic mean of R and P. Higher is better. Range 0 - 1"+\
          "\n"+\
          "F2 = Same as F1 Score, but put higher weights on R (recall). Range 0 - 1"+\
          "\n"+\
          "AUC = Area under precision recall curve. Higher is better. Range 0 - 1"+\
          "\n"+\
          "Topt = Optimal confidence threshold at maximum F1 score"+\
          "\n"+\
          "IOU = Intersection over Union, as higher the better the centering. Range 0 - 1"+\
          "\n"+\
          table.table+\
          "\n"+\
          "##########################"+\
          "\n"+\
          "Best confidence threshold ( -t ) according F1 statistic: "+ str(best_f1_thresh)+\
          "\n"+\
          "Best confidence threshold ( -t ) according F2 statistic: "+ str(best_f2_thresh)+\
          "\n"+\
          "##########################"

    print(out_res)

    #
    # Write result into .html
    #
    import codecs
    from ansi2html import Ansi2HTMLConverter
    conv = Ansi2HTMLConverter()
    html = conv.convert(out_res)

    output_file = codecs.open(path_output, "w",
                              encoding="ascii",
                              errors="xmlcharrefreplace"
                              )
    output_file.write(html)
    print("Wrote evaluation results to",path_output)

def get_num_fine_tune_layers(path_weights):
    import h5py
    with h5py.File(path_weights, mode='r') as f:
        try:
            num_free_layers = int(list(f["num_free_layers"])[0])
        except KeyError:
            num_free_layers = None
    return num_free_layers

def get_anchors(config, path_weights):
    is_single_anchor = config["model"]["anchors"] == 2
    grid_w, grid_h = config_tools.get_gridcell_dimensions(config)
    anchors = None
    import h5py
    with h5py.File(path_weights, mode='r') as f:
        try:
            anchors = list(f["anchors"])
        except KeyError:
            anchors = None

    if is_single_anchor and anchors is None:
        num_patches = 1
        if "num_patches" in config["model"]:
            num_patches = config["model"]["num_patches"]

        img_width = 4096.0 / num_patches
        img_height = 4096.0 / num_patches
        cell_w = img_width / grid_w
        cell_h = img_height / grid_h

        box_width, box_height = config_tools.get_box_size(config)

        anchor_width = 1.0 * box_width / cell_w
        anchor_height = 1.0 * box_height / cell_h
        anchors = [anchor_width, anchor_height]
    elif anchors is None:
        num_patches = 1
        if "num_patches" in config["model"]:
            num_patches = config["model"]["num_patches"]
        img_width = 4096.0 / num_patches
        img_height = 4096.0 / num_patches
        cell_w = img_width / grid_w
        cell_h = img_height / grid_h
        anchors = np.array(config["model"]["anchors"], dtype=float)
        anchors[::2] = anchors[::2] / cell_w
        anchors[1::2] = anchors[1::2] / cell_h

    return anchors


def prepare_datasets(
    path_config, path_weights, path_boxfiles, path_runfile, path_images
):
    datasets = {}

    if path_runfile:
        # Read run file
        with open(path_runfile) as runfile_buffer:
            runfile = json.load(runfile_buffer)
        images = runfile["run"]["valid_images"]
        annot = runfile["run"]["valid_annot"]

        for i in range(len(images)):
            imgpath = images[i]
            boxpath = annot[i]

            if os.stat(boxpath).st_size == 0:
                continue
            last_folder = os.path.basename(os.path.dirname(imgpath))
            if last_folder in datasets:
                box_files = datasets[last_folder]["boxes"]
                img_files = datasets[last_folder]["images"]
            else:
                box_files = []
                img_files = []
                datasets[last_folder] = {}
                datasets[last_folder]["boxes"] = box_files
                datasets[last_folder]["images"] = img_files
            box_files.append(boxpath)
            img_files.append(imgpath)
    else:
        # Scan all box filenames
        for root, directories, filenames in os.walk(path_boxfiles):
            if len(filenames) == 0:
                continue
            last_folder = os.path.basename(root)
            if last_folder.startswith(("empty", ".")):
                # Ignore empty grids
                continue

            box_files = []
            for filename in filenames:
                if filename.endswith(("box")) and not filename.startswith("."):
                    box_files.append(os.path.join(root, filename))

            if len(box_files) >= 1:
                if last_folder in datasets:
                    datasets[last_folder]["boxes"].append(box_files)
                else:
                    dataset = {"boxes": box_files}
                    datasets[last_folder] = dataset

        # Find all images
        img_files = []
        for root, directories, filenames in os.walk(path_images):
            last_folder = os.path.basename(root)
            if last_folder.startswith("empty"):
                # Ignore empty grids
                continue

            for filename in filenames:
                if filename.endswith(
                    ("jpg", "png", "tiff", "tif", "mrc")
                ) and not filename.startswith("."):
                    img_files.append(os.path.join(root, filename))

        # Find to boxifles corresponding images
        for foldername, dataset in datasets.items():
            boxfiles = dataset["boxes"]
            imgfiles = []
            for boxfile in boxfiles:
                filename = os.path.splitext(os.path.basename(boxfile))[0]
                lastfolder = os.path.basename(os.path.dirname(boxfile))
                search_path = os.path.join(lastfolder, filename)
                corresponding_img_path = difflib.get_close_matches(
                    search_path, img_files, n=1, cutoff=0
                )[0]
                print("Assign", boxfile, "to", corresponding_img_path)
                imgfiles.append(corresponding_img_path)
            dataset["images"] = imgfiles

    # Read config
    with open(path_config) as config_buffer:
        config = json.load(config_buffer)

    # Configure model
    ###############################
    #   Make the model
    ###############################
    depth = 1
    num_gpus = 1

    backend_weights = None

    anchors = get_anchors(config, path_weights)
    num_free_layers = get_num_fine_tune_layers(path_weights)
    if num_free_layers is None:
        num_free_layers = 2
    if "backend_weights" in config["model"]:
        backend_weights = config["model"]["backend_weights"]

    yolo = YOLO(
        architecture=config["model"]["architecture"],
        input_size=config["model"]["input_size"],
        input_depth=depth,
        labels=["particle"],
        max_box_per_image=config["model"]["max_box_per_image"],
        anchors=anchors,
        backend_weights=backend_weights,
        num_fine_tune_layers=num_free_layers
    )

    ###############################
    #   Load trained weights
    ###############################
    try:
        yolo.load_weights(path_weights,num_layers=num_free_layers)
    except ValueError as e:
        print(e)
        # print(traceback.format_exc())
        import sys
        sys.exit(
            "Seems that the architecture in your config (-c parameter) "
            "does not fit the model weights (-w parameter)"
        )
    # USE MULTIGPU
    if num_gpus > 1:
        parallel_model = multi_gpu_model(yolo.model, gpus=num_gpus)
        yolo.model = parallel_model

    # Get picks
    for foldername, dataset in datasets.items():
        # Get image dimensions of current dataset
        width, height = imagereader.read_width_height(dataset["images"][0])

        # Get box size
        one_box_file = read_box(dataset["boxes"][0], height)
        box_size = one_box_file[0]["w"]
        # Adapt the anchor size in config file.
        config["model"]["particle_diameter"] = box_size
        # if is_single_anchor:
        #     anchor_width = 1.0 * box_size / cell_w
        #     anchor_height = 1.0 * box_size / cell_h
        #     anchors = [anchor_width, anchor_height]
        #     yolo.anchors = anchors

        num_patches = config_tools.get_number_patches(config)

        # Get overlap patches
        overlap_patches = 0
        if "overlap_patches" in config["model"]:
            overlap_patches = int(config["model"]["overlap_patches"])
        elif not len(config["model"]["anchors"]) > 2:
            overlap_patches = config["model"]["anchors"][0]

        # Predict
        predict_results = predict.do_prediction(
            config_path=path_config,
            weights_path=path_weights,
            num_patches=num_patches,
            input_path=dataset["images"],
            obj_threshold=0.0,
            config_pre=config,
            overlap=overlap_patches,
            yolo=yolo,
        )

        # Rescale to image dimensions
        for boxes_in_micrograph in predict_results:
            image_width = boxes_in_micrograph["img_width"]
            image_height = boxes_in_micrograph["img_height"]
            for box in boxes_in_micrograph["boxes"]:
                x_ll = int((box.x - box.w / 2) * image_height)  # lower left
                y_ll = int(
                    image_width - box.y * image_width - box.h / 2 * image_width
                )  # lower right
                boxheight_in_pxl = int(box.h * image_width)
                boxwidth_in_pxl = int(box.w * image_height)

                delta_boxsize_height = box_size - boxheight_in_pxl
                delta_boxsize_width = box_size - boxwidth_in_pxl
                x_ll = x_ll - delta_boxsize_height / 2
                y_ll = y_ll - delta_boxsize_width / 2
                boxsize_in_pxl = box_size

                box.x = x_ll + boxsize_in_pxl / 2
                box.y = y_ll + boxsize_in_pxl / 2
                box.w = boxsize_in_pxl
                box.h = boxsize_in_pxl

        dataset["picks"] = predict_results

    return datasets


def calculate_all_statistics(datasets):

    arg_list = []

    for foldername, dataset in datasets.items():
        arg_list.append((foldername, dataset))

    pool = multiprocessing.Pool()
    result_dicts = pool.starmap(calculate_statics_dataset, arg_list)
    pool.close()
    pool.join()

    k = 0
    for foldername, dataset in datasets.items():
        dataset["results"] = result_dicts[k]
        """
        result_rows = zip(result_dicts[k]["RECALL_LIST"], result_dicts[k]["PRECISION_LIST"])

        import csv
        with open("recall_precision_"+foldername+".csv", "w") as f:
            writer = csv.writer(f, delimiter=';')
            for row in result_rows:

                writer.writerow(row)

        f.close()
        """
        k = k + 1


def calculate_statics_dataset(foldername, dataset):
    arange = np.arange(start=0, stop=1.001, step=0.001)
    result_dict = {}
    maxf1 = 0.0
    maxf1_thresh = 0.0
    maxf2_thresh = 0.0
    maxf2 = 0.0
    recall_list = []
    precision_list = []
    f1_list = []
    f2_list = []
    sum_picked_list = []

    for thresh in arange:
        thresh = round(thresh, 3)
        res = get_precision_recall_f1(dataset, thresh)
        if res is not None:
            result_dict[str(thresh)] = res
            precision = result_dict[str(thresh)][1]
            recall = result_dict[str(thresh)][2]
            sum_picked = result_dict[str(thresh)][4]
            f1 = result_dict[str(thresh)][0]
            f2 = result_dict[str(thresh)][5]
            precision_list.append(precision)
            recall_list.append(recall)
            f1_list.append(f1)
            f2_list.append(f2)
            sum_picked_list.append(sum_picked)
            # if len(recall_list) > 1:
            #    area.append((recall_list[-2] - recall_list[-1]) * precision_list[-1])

            if f1 > maxf1:
                maxf1 = f1
                maxf1_thresh = thresh
            if f2 > maxf2:
                maxf2 = f2
                maxf2_thresh = thresh

    # AUC Calculation
    precision_array = np.array(precision_list)
    recall_array = np.array(recall_list)

    # Interpolate first point
    recall_array = np.insert(recall_array, len(recall_array), 0)
    precision_array = np.insert(
        precision_array, len(precision_array), precision_array[len(precision_array) - 1]
    )

    # Sort
    sorted_index = np.argsort(recall_array)[::-1]
    precision_array = precision_array[sorted_index]
    recall_array = recall_array[sorted_index]

    area = 0
    for i in range(1, len(recall_array)):
        val = (recall_array[i - 1] - recall_array[i]) * precision_array[i]
        area = area + val

    result_dict["AUC"] = area
    result_dict["MAX_F1_THRESH"] = maxf1_thresh
    result_dict["MAX_F2_THRESH"] = maxf2_thresh
    result_dict["THRESH_LIST"] = arange
    result_dict["RECALL_LIST"] = recall_array.tolist()
    result_dict["PRECISION_LIST"] = precision_array.tolist()
    result_dict["F1_LIST"] = f1_list
    result_dict["F2_LIST"] = f2_list
    result_dict["PICKED_LIST"] = sum_picked_list
    return result_dict


def get_precision_recall_f1(dataset, confidence_threshold):

    results_per_threshold = {}

    f1_sum = 0
    f2_sum = 0
    precision_sum = 0
    recall_sum = 0
    mean_iou_sum = 0
    iou_tresh = 0.5
    picks = dataset["picks"]

    boxes = dataset["boxes"]
    images = dataset["images"]
    sum_picked = 0
    images_with_particles = 0
    for boxes_in_mic in picks:
        TP = 0
        MEAN_IOU = 0
        gd_box_remove = []
        picked_boxes_index_remove = []

        picked_boxes = boxes_in_mic["boxes"]
        img_path = boxes_in_mic["img_path"]
        img_name_without_suffix = os.path.splitext(os.path.basename(img_path))[0]
        # ground_truth_boxes_path = difflib.get_close_matches(img_name_without_suffix, boxes, n=1, cutoff=0)[0]

        cand_list = [
            i
            for i in boxes
            if os.path.splitext(os.path.basename(i))[0] in img_name_without_suffix
        ]
        try:
            cand_list_no_fileextension = list(map(os.path.basename, cand_list))
            ground_truth_boxes_path = difflib.get_close_matches(
                img_name_without_suffix, cand_list_no_fileextension, n=1, cutoff=0
            )[0]
            ground_truth_boxes_path = cand_list[
                cand_list_no_fileextension.index(ground_truth_boxes_path)
            ]
        except IndexError as e:
            print("Cannot find corresponding boxfile for ", img_path)
            raise e

        # Read ground truth boxes:
        _, height = imagereader.read_width_height(img_path)
        ground_truth_boxes = read_box(ground_truth_boxes_path, height)
        picked_boxes_filtered = [
            x for x in picked_boxes if x.get_score() > confidence_threshold
        ]
        sum_picked = sum_picked + len(picked_boxes_filtered)
        # Ground Truth boxes to array
        gd_boxes_data = np.empty(shape=(len(ground_truth_boxes), 5))
        for i, gdbox in enumerate(ground_truth_boxes):
            gd_boxes_data[i, 0] = gdbox["x"]
            gd_boxes_data[i, 1] = gdbox["y"]
            gd_boxes_data[i, 2] = gdbox["w"]
            gd_boxes_data[i, 3] = gdbox["h"]
            gd_boxes_data[i, 4] = i

        # Picked boxes to array
        picked_boxes_data = np.empty(shape=(len(picked_boxes_filtered), 5))
        for i, pbox in enumerate(picked_boxes_filtered):
            picked_boxes_data[i, 0] = pbox.x
            picked_boxes_data[i, 1] = pbox.y
            picked_boxes_data[i, 2] = pbox.w
            picked_boxes_data[i, 3] = pbox.h
            picked_boxes_data[i, 4] = i
        # print("GD LEN",len(gd_boxes_data))
        # print("PICKED LEN", len(picked_boxes_data), "Tresh",confidence_threshold, "img",img_path)
        for i, gd_box in enumerate(ground_truth_boxes):

            # Run fast methods here
            if len(picked_boxes_data) > 0:
                box_i = gd_boxes_data[i, :]
                boxes_i_rep = np.array([box_i] * len(picked_boxes_data))
                ious = bbox_iou_vec(boxes_i_rep, picked_boxes_data)

                # Delete that
                # ious = calc_ious(gd_box,picked_boxes_filtered)
                max_iou = 0
                p_box_index = 0
                if len(ious) > 0:
                    p_box_index = np.argmax(ious)
                    max_iou = ious[p_box_index]

                # print max_iou
                if max_iou > iou_tresh:
                    gd_box_remove.append(gd_box)
                    # p_box_index = ious.index(max_iou)
                    picked_boxes_index_remove.append(p_box_index)
                    TP = TP + 1
                    MEAN_IOU += max_iou
        if len(picked_boxes_filtered) > 0:
            FP = len(picked_boxes_filtered) - len(picked_boxes_index_remove)
            FN = len(ground_truth_boxes) - len(gd_box_remove)
            if (TP + FP) == 0:
                precision = 0
            else:
                precision = 1.0 * TP / (TP + FP)

            recall = 1.0 * TP / (TP + FN)
            if (precision + recall) == 0:
                F1 = 0
                F2 = 0
            else:
                F1 = 2.0 * precision * recall / (precision + recall)
                F2 = 5.0 * precision * recall / (4 * precision + recall)
            f1_sum += F1
            f2_sum += F2
            precision_sum += precision
            recall_sum += recall
            mean_iou_sum += MEAN_IOU / (TP + 0.001)

            images_with_particles = images_with_particles + 1

    if images_with_particles == 0:
        return None

    # --------------------------------------------#
    f1_avg = f1_sum / images_with_particles
    f2_avg = f2_sum / images_with_particles
    precision_avg = precision_sum / images_with_particles
    recall_avg = recall_sum / images_with_particles
    mean_iou_sum = mean_iou_sum / images_with_particles
    return (f1_avg, precision_avg, recall_avg, mean_iou_sum, sum_picked, f2_avg)


def calc_stats_single_mic(boxes_in_mic, boxes, iou_tresh, confidence_threshold):
    TP = 0
    MEAN_IOU = 0
    gd_box_remove = []
    picked_boxes_index_remove = []

    picked_boxes = boxes_in_mic["boxes"]
    img_path = boxes_in_mic["img_path"]
    img_name_without_suffix = os.path.splitext(os.path.basename(img_path))[0]
    # ground_truth_boxes_path = difflib.get_close_matches(img_name_without_suffix, boxes, n=1, cutoff=0)[0]

    cand_list = [
        i
        for i in boxes
        if os.path.splitext(os.path.basename(i))[0] in img_name_without_suffix
    ]
    try:
        cand_list_no_fileextension = list(map(os.path.basename, cand_list))
        ground_truth_boxes_path = difflib.get_close_matches(
            img_name_without_suffix, cand_list_no_fileextension, n=1, cutoff=0
        )[0]
        ground_truth_boxes_path = cand_list[
            cand_list_no_fileextension.index(ground_truth_boxes_path)
        ]
    except IndexError as e:
        print("Cannot find corresponding boxfile for ", img_path)
        raise e

    # Read ground truth boxes:
    _, height = imagereader.read_width_height(img_path)
    ground_truth_boxes = read_box(ground_truth_boxes_path, height)
    picked_boxes_filtered = [x for x in picked_boxes if x.c > confidence_threshold]

    # Ground Truth boxes to array
    gd_boxes_data = np.empty(shape=(len(ground_truth_boxes), 5))
    for i, gdbox in enumerate(ground_truth_boxes):
        gd_boxes_data[i, 0] = gdbox["x"]
        gd_boxes_data[i, 1] = gdbox["y"]
        gd_boxes_data[i, 2] = gdbox["w"]
        gd_boxes_data[i, 3] = gdbox["h"]
        gd_boxes_data[i, 4] = i

    # Picked boxes to array
    picked_boxes_data = np.empty(shape=(len(picked_boxes_filtered), 5))
    for i, pbox in enumerate(picked_boxes_filtered):
        picked_boxes_data[i, 0] = pbox.x
        picked_boxes_data[i, 1] = pbox.y
        picked_boxes_data[i, 2] = pbox.w
        picked_boxes_data[i, 3] = pbox.h
        picked_boxes_data[i, 4] = i
    # print("GD LEN",len(gd_boxes_data))
    # print("PICKED LEN", len(picked_boxes_data))
    for i, gd_box in enumerate(ground_truth_boxes):

        # Run fast methods here
        if len(picked_boxes_data) > 0:
            box_i = gd_boxes_data[i, :]
            boxes_i_rep = np.array([box_i] * len(picked_boxes_data))
            ious = bbox_iou_vec(boxes_i_rep, picked_boxes_data)

            # Delete that
            # ious = calc_ious(gd_box,picked_boxes_filtered)
            max_iou = 0
            p_box_index = 0
            if len(ious) > 0:
                p_box_index = np.argmax(ious)
                max_iou = ious[p_box_index]

            # print max_iou
            if max_iou > iou_tresh:
                gd_box_remove.append(gd_box)
                # p_box_index = ious.index(max_iou)
                picked_boxes_index_remove.append(p_box_index)
                TP = TP + 1
                MEAN_IOU += max_iou
    FP = len(picked_boxes_filtered) - len(picked_boxes_index_remove)
    FN = len(ground_truth_boxes) - len(gd_box_remove)

    if (TP + FP) == 0:
        precision = 0
    else:
        precision = 1.0 * TP / (TP + FP)

    recall = 1.0 * TP / (TP + FN)
    if (precision + recall) == 0:
        F1 = 0
    else:
        F1 = 2.0 * precision * recall / (precision + recall)
    return (F1, precision, recall, MEAN_IOU / (TP + 0.001))


def calc_ious(ground_truth_box, picked_boxes):
    return [bbox_iou(ground_truth_box, p_box) for p_box in picked_boxes]


def read_box(path, image_height):
    boxes = []
    # with open(path, 'r') as boxfile:
    boxreader = np.atleast_2d(
        np.genfromtxt(path)
    )  # csv.reader(boxfile, delimiter='\t',
    #          quotechar='|', quoting=csv.QUOTE_NONE)

    try:
        for box in boxreader:
            box_d = {}
            width = int(float(box[2]))
            height = int(float(box[3]))

            box_d["x"] = int(float(box[0])) + width / 2
            box_d["y"] = int(float(box[1])) + height / 2
            box_d["w"] = width
            box_d["h"] = height
            box_d["filename"] = path
            boxes.append(box_d)
    except Exception as e:
        print("ReadBox error. Path:", path)
        print(e)

    return boxes


def read_boxes(box_dir, image_height):
    onlyfiles = [
        f for f in os.listdir(box_dir) if os.path.isfile(os.path.join(box_dir, f))
    ]
    boxes = []
    for file in onlyfiles:
        if file.endswith(".box"):
            path = os.path.join(box_dir, file)
            box = read_box(path, image_height)
            if box is not None:
                boxes.append(box)
    return boxes


def interval_overlap(interval_a, interval_b):
    x1, x2 = interval_a
    x3, x4 = interval_b

    if x3 < x1:
        if x4 < x1:
            return 0
        else:
            return min(x2, x4) - x1
    else:
        if x2 < x3:
            return 0
        else:
            return min(x2, x4) - x3


def bbox_iou_vec(boxesA, boxesB):
    x1_min = boxesA[:, 0] - boxesA[:, 2] / 2
    x1_max = boxesA[:, 0] + boxesA[:, 2] / 2
    y1_min = boxesA[:, 1] - boxesA[:, 3] / 2
    y1_max = boxesA[:, 1] + boxesA[:, 3] / 2

    x2_min = boxesB[:, 0] - boxesB[:, 2] / 2
    x2_max = boxesB[:, 0] + boxesB[:, 2] / 2
    y2_min = boxesB[:, 1] - boxesB[:, 3] / 2
    y2_max = boxesB[:, 1] + boxesB[:, 3] / 2
    intersect_w = interval_overlap_vec(x1_min, x1_max, x2_min, x2_max)
    intersect_h = interval_overlap_vec(y1_min, y1_max, y2_min, y2_max)
    intersect = intersect_w * intersect_h
    union = boxesA[:, 2] * boxesA[:, 3] + boxesB[:, 2] * boxesB[:, 3] - intersect
    return intersect / union


def interval_overlap_vec(x1_min, x1_max, x2_min, x2_max):
    intersect = np.zeros(shape=(len(x1_min)))
    condA = x2_min < x1_min
    condB = condA & (x2_max >= x1_min)
    intersect[condB] = np.minimum(x1_max[condB], x2_max[condB]) - x1_min[condB]
    condC = ~condA & (x1_max >= x2_min)
    intersect[condC] = np.minimum(x1_max[condC], x2_max[condC]) - x2_min[condC]

    return intersect


def bbox_iou(box1, box2):

    x1_min = box1["x"] - box1["w"] / 2
    x1_max = box1["x"] + box1["w"] / 2
    y1_min = box1["y"] - box1["h"] / 2
    y1_max = box1["y"] + box1["h"] / 2

    x2_min = box2.x - box2.w / 2
    x2_max = box2.x + box2.w / 2
    y2_min = box2.y - box2.h / 2
    y2_max = box2.y + box2.h / 2

    intersect_w = interval_overlap([x1_min, x1_max], [x2_min, x2_max])
    intersect_h = interval_overlap([y1_min, y1_max], [y2_min, y2_max])

    intersect = intersect_w * intersect_h

    union = box1["w"] * box1["h"] + box2.w * box2.h - intersect

    return float(intersect) / union


class picking_result_folder:
    def __init__(self, foldername):
        self.foldername = foldername
        self.boxes = {}

    def add_boxes(self, image_name, boxes_dict):
        self.boxes[image_name] = boxes_dict


if __name__ == "__main__":
    _main_()



from gooey import Gooey
import argparse

@Gooey
def _main_():
    parent = argparse.ArgumentParser(description="My program")
    parent.add_argument("-g","--global",help="A global variable")
    subparsers = parent.add_subparsers(help="")
    parser_child_a = subparsers.add_parser("childa", help="childa help", parents=[parent],
                                         add_help=False)
    parser_child_a.add_argument("-a",help="a option")

    parser_child_b = subparsers.add_parser("childb", help="childb help", parents=[parent],
                                           add_help=False)
    parser_child_b.add_argument("-b", help="b option")

    parent.parse_args()

if __name__ == "__main__":
    _main_()


#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

import tensorflow as tf
from keras.layers import Lambda, concatenate
from keras.models import Model


def my_multi_gpu_model(model, gpus):
    if isinstance(gpus, (list, tuple)):
        num_gpus = len(gpus)
        target_gpu_ids = gpus
    else:
        num_gpus = gpus
        target_gpu_ids = range(num_gpus)

    def get_slice(data, i, parts):
        shape = tf.shape(data)
        batch_size = shape[:1]
        input_shape = shape[1:]
        step = batch_size // parts
        if i == num_gpus - 1:
            size = batch_size - step * i
        else:
            size = step
        size = tf.concat([size, input_shape], axis=0)
        stride = tf.concat([step, input_shape * 0], axis=0)
        start = stride * i
        return tf.slice(data, start, size)

    all_outputs = []
    for i in range(len(model.outputs)):
        all_outputs.append([])

    # Place a copy of the model on each GPU,
    # each getting a slice of the inputs.
    for i, gpu_id in enumerate(target_gpu_ids):
        with tf.device("/gpu:%d" % gpu_id):
            with tf.name_scope("replica_%d" % gpu_id):
                inputs = []
                # Retrieve a slice of the input.
                for x in model.inputs:
                    input_shape = tuple(x.get_shape().as_list())[1:]
                    slice_i = Lambda(
                        get_slice,
                        output_shape=input_shape,
                        arguments={"i": i, "parts": num_gpus},
                    )(x)
                    inputs.append(slice_i)

                # Apply model on slice
                # (creating a model replica on the target device).
                outputs = model(inputs)
                if not isinstance(outputs, list):
                    outputs = [outputs]

                # Save the outputs for merging back together later.
                for o in range(len(outputs)):
                    all_outputs[o].append(outputs[o])

    # Merge outputs on CPU.
    with tf.device("/cpu:0"):
        merged = []
        for name, outputs in zip(model.output_names, all_outputs):
            merged.append(concatenate(outputs, axis=0, name=name))
        return Model(model.inputs, merged)



#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

from keras.callbacks import ModelCheckpoint


class ExtendedModelCheckpoint(ModelCheckpoint):

    def set_fine_tune(self, fine_tune, num_free_layers):
        self.fine_tune = fine_tune
        self.num_free_layers = num_free_layers

    def set_anchors(self, anchors):
        self.anchors = anchors


    def on_epoch_end(self, epoch, logs=None):
        update = False
        current = logs.get(self.monitor)
        if current is not None:
            if self.monitor_op(current, self.best):
                update = True
        super(ExtendedModelCheckpoint, self).on_epoch_end(epoch, logs)

        if update:
            #############################################
            # Save meta data about the model
            #############################################
            import h5py
            with h5py.File(self.filepath, mode='r+') as f:
                f["anchors"] = self.anchors
                if self.fine_tune:
                    f["num_free_layers"] = [self.num_free_layers]




#! /usr/bin/env python

#
# COPYRIGHT
#
# All contributions by Ngoc Anh Huyn:
# Copyright (c) 2017, Ngoc Anh Huyn.
# All rights reserved.
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------

from __future__ import print_function

import argparse
import json
import multiprocessing
import os
import time

import numpy as np

import cryolo.config_tools as config_tools
import cryolo.imagereader as imagereader
import cryolo.lowpass as lowpass
from cryolo.preprocessing import parse_annotation2
from . import utils
from gooey import Gooey, GooeyParser


os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
try:
    os.environ["CUDA_VISIBLE_DEVICES"]
except KeyError:
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"


def create_parser(parser):
    required_group = parser.add_argument_group(
        "Required arguments", "These options are mandatory to run crYOLO train"
    )

    required_group.add_argument(
        "-c",
        "--conf",
        required=True,
        help="path to configuration file",
        widget="FileChooser",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("json")',
                "message": "File has to end with .json!",
            },
            "wildcard":".json"
        }
    )

    required_group.add_argument(
        "-w",
        "--warmup",
        type=int,
        required=True,
        default=5,
        help="Number of warmup epochs. Set it to zero if you fine tune a model.",
    )

    optional_group = parser.add_argument_group(
        "Optional arguments", "These options are optional to run crYOLO train"
    )

    optional_group.add_argument(
        "-g",
        "--gpu",
        default="",
        nargs="+",
        help="Specifiy which gpu(s) should be used. Multiple GPUs are separated by a whitespace",
    )

    optional_group.add_argument(
        "-nc",
        "--num_cpu",
        type=int,
        default=-1,
        help="Number of CPUs used during training. By default it will use half of the available CPUs.",
    )

    optional_group.add_argument(
        "--gpu_fraction",
        type=float,
        default=1.0,
        help="Specify the fraction of memory per GPU used by crYOLO during training. Only values between 0.0 and 1.0 are allowed.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "-e",
        "--early",
        default=10,
        type=int,
        help="Early stop patience. If the validation loss did not improve longer than the early stop patience, "
        "the training is stopped.",
    )


    optional_group.add_argument(
        "--fine_tune",
        action="store_true",
        default=False,
        help="Set it to true if you only want to use the fine tune mode.",
    )

    optional_group.add_argument(
        "-lft",
        "--layers_fine_tune",
        default=2,
        type=int,
        help="Layers to be trained when using fine tuning.",
    )

    deexsp_group = parser.add_argument_group(
        "Deprecated/Experimental/Special ",
        "Contains either deprecated / experimental or very special options."
    )

    deexsp_group.add_argument(
        "--seed",
        type=int,
        default=10,
        help="Seed for random number generator. Mainly influences selection of validation images. Should be the same during different training runs!",
    )

    deexsp_group.add_argument(
        "--warm_restarts",
        action="store_true",
        help="Use warm restarts and cosine annealing during training",
    )

    deexsp_group.add_argument(
        "--skip_augmentation",
        action="store_true",
        default=False,
        help="Use it if you want to deactivate data augmentation during training.",
    )

# Reset Keras Session
def reset_keras(gpu_fraction=1):
    from keras.backend.tensorflow_backend import set_session
    from keras.backend.tensorflow_backend import clear_session
    from keras.backend.tensorflow_backend import get_session
    import gc
    import tensorflow
    sess = get_session()
    clear_session()
    sess.close()
    sess = get_session()

    print(gc.collect()) # if it's done something you should see a number being outputted

    # use the same config as you used to create the session
    config = tensorflow.ConfigProto()
    config.gpu_options.per_process_gpu_memory_fraction = gpu_fraction
    #config.gpu_options.visible_device_list = "0"
    set_session(tensorflow.Session(config=config))


def get_parser():
    parser = GooeyParser(
        description="Train crYOLO model on any datasett",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    create_parser(parser)
    return parser

def get_model(config, depth, anchors, experimental_loss, do_fine_tune, num_fine_tune_layers, num_gpus, pretrained_weights):
    ###############################
    #   Construct the model
    ###############################
    backend_weights = None
    if "backend_weights" in config["model"]:
        backend_weights = config["model"]["backend_weights"]
    from cryolo.frontend import YOLO



    #import tensorflow as tf
    #graph = tf.Graph()
    #with graph.as_default():
    if num_gpus > 1:
        import tensorflow as tf

        with tf.device("/cpu:0"):
            yolo = YOLO(
                architecture=config["model"]["architecture"],
                input_size=config["model"]["input_size"],
                input_depth=depth,
                labels=config["model"]["labels"],
                max_box_per_image=config["model"]["max_box_per_image"],
                anchors=anchors,
                backend_weights=backend_weights,
                experimental_loss=experimental_loss,
                pretrained_weights=pretrained_weights,
                fine_tune=do_fine_tune,
                num_fine_tune_layers=num_fine_tune_layers,
            )
    else:
        yolo = YOLO(
            architecture=config["model"]["architecture"],
            input_size=config["model"]["input_size"],
            input_depth=depth,
            labels=config["model"]["labels"],
            max_box_per_image=config["model"]["max_box_per_image"],
            anchors=anchors,
            backend_weights=backend_weights,
            experimental_loss=experimental_loss,
            pretrained_weights=pretrained_weights,
            fine_tune=do_fine_tune,
            num_fine_tune_layers=num_fine_tune_layers,
        )

    # print a summary of the whole model
    yolo.model.summary()

    return yolo

def _main_():
    try:
        multiprocessing.set_start_method("spawn")
    except RuntimeError:
        print("Ignore set start method")

    import sys

    # if sys.argv[1] == "gui":
    if len(sys.argv) >= 2:
        if not "--ignore-gooey" in sys.argv:
            sys.argv.append("--ignore-gooey")

    # r'^\d+ particles are found in .* \( (\d+) % \)$'
    kwargs = {"terminal_font_family": "monospace",
              "richtext_controls": True}
    Gooey(
        main,
        program_name="crYOLO train",
        image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        **kwargs
    )()


def main(args=None):
    print("START METHOD:", multiprocessing.get_start_method())

    if args is None:
        parser = get_parser()
        args = parser.parse_args()

    if isinstance(args.gpu, list):
        if len(args.gpu) == 1:
            str_gpus = args.gpu[0].strip().split(" ")
        else:
            str_gpus = [str(entry) for entry in args.gpu]
        num_gpus = len(str_gpus)
        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
    else:
        num_gpus = 1
        if args.gpu != -1 and len(args.gpu) > 0:
            str_gpus = str(args.gpu)
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)

    if args.gpu_fraction < 1.0 and args.gpu_fraction > 0.0:
        import tensorflow as tf
        from keras.backend.tensorflow_backend import set_session

        config = tf.ConfigProto()

        config.gpu_options.per_process_gpu_memory_fraction = args.gpu_fraction

        set_session(tf.Session(config=config))

    config_path = args.conf

    early_stop = int(args.early)

    warm_restarts = args.warm_restarts

    do_augmentation = not args.skip_augmentation

    do_fine_tune = args.fine_tune
    num_fine_tune_layers = args.layers_fine_tune

    num_cpus = args.num_cpu
    seed = args.seed
    if args.warmup is not None:
        warmup_epochs = int(args.warmup)
        print("Read warmup by argument")
    else:
        if config["train"]["warmup_epochs"] is not None:
            warmup_epochs = config["train"]["warmup_epochs"]
            print("Read warmup by config")
        else:
            warmup_epochs = 0
            print("Set warmup to zero")

    if args.warmup is not None and do_fine_tune:
        warmup_epochs = 0
        print("Ignore warmup, as fine tuning is activated.")

    if early_stop < warmup_epochs:
        early_stop = warmup_epochs

    experimental_loss = False

    #
    # Read and adapt configuration
    #

    with open(config_path) as config_buffer:
        try:
            config = json.loads(config_buffer.read())
        except json.JSONDecodeError:
            print(
                "Your configuration file seems to be corruped. Please check if it is valid."
            )
        if config["model"]["input_size"] % 32 > 0:
            new_input_size = int(32 * round(float(config["model"]["input_size"]) / 32))
            print(
                "You input size is not a multiple of 32. Round it to the next multiple of 32:",
                new_input_size,
            )
            config["model"]["input_size"] = new_input_size
    num_patches = config_tools.get_number_patches(config)
    experimental_loss = False

    if early_stop < warmup_epochs:
        early_stop = warmup_epochs

    grid_w, grid_h = config_tools.get_gridcell_dimensions(config)

    ###############################
    #   Parse the annotations
    ###############################

    # parse annotations of the training set

    train_imgs, train_labels = parse_annotation2(
        ann_dir=config["train"]["train_annot_folder"],
        img_dir=config["train"]["train_image_folder"],
        grid_dims=(grid_w, grid_h, num_patches),
        anchor_size=int(config["model"]["anchors"][0]),
    )
    config["model"]["labels"] = ["particle"]

    # parse annotations of the validation set, if any, otherwise split the training set
    if os.path.exists(config["valid"]["valid_annot_folder"]):
        valid_imgs, valid_labels = parse_annotation2(
            config["valid"]["valid_annot_folder"],
            config["valid"]["valid_image_folder"],
            (grid_w, grid_h, num_patches),
        )

        if (
            len(valid_imgs) == 0
            or len(valid_labels) == 0
            or len(valid_labels) != len(valid_imgs)
        ):
            if len(valid_imgs) == 0:
                print(
                    "No validation images were found. Invalid validation configuration. Check your config file."
                )
            if len(valid_labels) == 0:
                print(
                    "No validation labels were found. Invalid validation configuration. Check your config file."
                )
    else:
        np.random.seed(seed)
        train_valid_split = int(0.8 * len(train_imgs))
        np.random.shuffle(train_imgs)
        valid_imgs = train_imgs[train_valid_split:]
        train_imgs = train_imgs[:train_valid_split]
        print("Validation set:")
        print([item["filename"] for item in valid_imgs])

    #####################################
    # Write runfile
    #####################################

    valid_imgs_paths = [item["filename"] for item in valid_imgs]
    valid_annot_paths = [item["boxpath"] for item in valid_imgs]
    train_imgs_paths = [item["filename"] for item in train_imgs]
    train_annot_paths = [item["boxpath"] for item in train_imgs]

    runjson = {}
    runjson["run"] = {}
    runjson["run"]["valid_images"] = valid_imgs_paths
    runjson["run"]["valid_annot"] = valid_annot_paths
    runjson["run"]["train_images"] = train_imgs_paths
    runjson["run"]["train_annot"] = train_annot_paths

    if not os.path.exists("runfiles/"):
        os.mkdir("runfiles/")
    timestr = time.strftime("%Y%m%d-%H%M%S")
    with open("runfiles/run_" + timestr + ".json", "w") as outfile:
        json.dump(runjson, outfile, ensure_ascii=False, indent=4)

    ##############################
    # Filter the data
    ##############################

    if "filter" in config["model"]:
        filter_options = config["model"]["filter"]
        if len(filter_options) > 2:
            model_path, overlap, nn_batch_size, filter_img_path = filter_options
            print("Filter data using noise2noise model: ", model_path)

            # Correct absolute filenames
            train_imgs_paths = [img["filename"] for img in train_imgs]

            filtered_paths = utils.filter_images_nn_dir(
                img_paths=train_imgs_paths,
                output_dir_filtered_imgs=filter_img_path,
                model_path=model_path,
                padding=overlap,
                batch_size=nn_batch_size,
            )

            # Update paths of newly filtered images
            for index, path in enumerate(filtered_paths):
                train_imgs[index]["filename"] = path

            valid_imgs_paths = [img["filename"] for img in valid_imgs]
            filtered_paths = utils.filter_images_nn_dir(
                img_paths=valid_imgs_paths,
                output_dir_filtered_imgs=filter_img_path,
                model_path=model_path,
                padding=overlap,
                batch_size=nn_batch_size,
            )
            # Update paths of newly filtered images
            for index, path in enumerate(filtered_paths):
                valid_imgs[index]["filename"] = path

        else:
            cutoff, filter_img_path = filter_options
            lowpass.filter_images(train_imgs, cutoff, filter_img_path, num_cpus)
            lowpass.filter_images(valid_imgs, cutoff, filter_img_path, num_cpus)

    overlap_labels = set(config["model"]["labels"]).intersection(
        set(train_labels.keys())
    )

    # Read first image and check the image depth.
    img_first = imagereader.image_read(train_imgs[0]["filename"])
    isgrey = False
    if len(img_first.shape) == 2:
        isgrey = True
    elif np.all(img_first[:, :, 0] == img_first[:, :, 1]) and np.all(
        img_first[:, :, 0] == img_first[:, :, 2]
    ):
        isgrey = True

    if isgrey:
        depth = 1
    else:
        depth = 3

    # As only on box size is expected, the anchor box size automatically

    if len(config["model"]["anchors"]) > 2:
        # general model is used
        img_width = 4096.0 / num_patches
        img_height = 4096.0 / num_patches
        cell_w = img_width / grid_w
        cell_h = img_height / grid_h
        anchors = np.array(config["model"]["anchors"], dtype=float)
        anchors[::2] = anchors[::2] / cell_w
        anchors[1::2] = anchors[1::2] / cell_h

    else:
        # specifc model is used
        img_width = float(img_first.shape[1]) / num_patches
        img_height = float(img_first.shape[0]) / num_patches

        cell_w = img_width / grid_w
        cell_h = img_height / grid_h
        box_width, box_height = config_tools.get_box_size(config)

        anchor_width = 1.0 * box_width / cell_w
        anchor_height = 1.0 * box_height / cell_h
        anchors = [anchor_width, anchor_height]

    if not config["train"]["log_path"]:
        log_path = "~/logs/"
    else:
        log_path = config["train"]["log_path"]
        if not os.path.exists(log_path):
            os.mkdir(log_path)

    # Get overlap patches
    overlap_patches = 0
    if "overlap_patches" in config["model"]:
        overlap_patches = int(config["model"]["overlap_patches"])
    elif not len(config["model"]["anchors"]) > 2:
        overlap_patches = config["model"]["anchors"][0]



    ###############################
    # Start the training process
    ###############################
    pretrained_weights = None
    if os.path.exists(config["train"]["pretrained_weights"]):
        pretrained_weights = config["train"]["pretrained_weights"]

    from keras import backend as K
    import tensorflow as tf
    while True:
            start = time.time()
            #graph = tf.Graph()
            yolo = get_model(config, depth, anchors, experimental_loss, do_fine_tune, num_fine_tune_layers, num_gpus, pretrained_weights)

            # USE MULTIGPU
            parallel_model = None
            if num_gpus > 1:
                from keras.utils import multi_gpu_model

                parallel_model = multi_gpu_model(yolo.model, gpus=num_gpus, cpu_merge=False)

            yolo.train(
                train_imgs=train_imgs,
                valid_imgs=valid_imgs,
                train_times=config["train"]["train_times"],
                valid_times=config["valid"]["valid_times"],
                nb_epoch=config["train"]["nb_epoch"],
                learning_rate=config["train"]["learning_rate"],
                batch_size=config["train"]["batch_size"],
                warmup_epochs=warmup_epochs,
                object_scale=config["train"]["object_scale"],
                no_object_scale=config["train"]["no_object_scale"],
                coord_scale=config["train"]["coord_scale"],
                class_scale=config["train"]["class_scale"],
                saved_weights_name=config["train"]["saved_weights_name"],
                debug=config["train"]["debug"],
                log_path=log_path,
                early_stop_thresh=early_stop,
                num_patches=num_patches,
                warm_restarts=warm_restarts,
                overlap_patches=overlap_patches,
                parallel_model=parallel_model,
                num_cpus=num_cpus,
                do_augmentation=do_augmentation,
            )

            end = time.time()
            if warmup_epochs > 0:
                print("Time elapsed for warmup:", (end - start))
            else:
                print("Time elapsed for training:", (end - start))

            if warmup_epochs == 0:
                break
            else:
                warmup_epochs = 0
                pretrained_weights = config["train"]["saved_weights_name"]
                reset_keras(args.gpu_fraction) # Only method that allows me to train crYOLO in a loop


if __name__ == "__main__":
    multiprocessing.set_start_method("spawn")
    _main_()



#
# COPYRIGHT
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

from __future__ import print_function

import difflib
import os

import mrcfile
import numpy as np
from PIL import Image

import cryolo.CoordsIO as CoordsIO


def parse_annotation2(ann_dir, img_dir, grid_dims=None, anchor_size=None):
    if not os.path.exists(ann_dir):
        import sys

        print(
            "Your annotation folder does not exists: ",
            ann_dir,
            " Check your config file!",
        )
        sys.exit()

    if not os.path.exists(img_dir):
        import sys

        print(
            "Your image folder does not exists: ", img_dir, " Check your config file!"
        )
        sys.exit()

    all_imgs = []
    seen_labels = {}
    img_files = []

    # Scan all image filenames
    for root, directories, filenames in os.walk(img_dir, followlinks=True):
        for filename in filenames:
            if filename.endswith(
                ("jpg", "png", "tiff", "tif", "mrc")
            ) and not filename.startswith("."):
                img_files.append(os.path.join(root, filename))

    # Read annotations
    img_names = list(map(os.path.basename, img_files))
    for root, directories, filenames in os.walk(ann_dir, followlinks=True):
        for ann in sorted(filenames):
            if ann.endswith((".box", ".txt", ".star")) and not ann.startswith("."):

                img = {"object": []}
                ann_without_ext = os.path.splitext(os.path.basename(ann))[0]

                cand_list = [i for i in img_names if ann_without_ext in i]
                try:
                    cand_list_no_fileextension = list(map(os.path.basename, cand_list))
                    corresponding_img_path = difflib.get_close_matches(
                        ann_without_ext, cand_list_no_fileextension, n=1, cutoff=0
                    )[0]
                    corresponding_img_path = cand_list[
                        cand_list_no_fileextension.index(corresponding_img_path)
                    ]

                except IndexError:
                    print("Cannot find corresponding image file for ", root, ann)
                    import sys

                    sys.exit()
                index_image = img_names.index(corresponding_img_path)
                img["filename"] = img_files[index_image]
                print(ann, " assigned to ", img["filename"])
                if img["filename"].endswith("mrc"):
                    with mrcfile.mmap(img["filename"], permissive=True) as mrc:
                        img_height = mrc.header.ny

                else:
                    im = Image.open(img["filename"])
                    _, img_height = [int(i) for i in im.size]
                boxpath = os.path.join(root, ann)
                img["boxpath"] = boxpath

                is_helicion_ptcl = is_helicon_with_particle_coords(boxpath)
                is_helicion_eman = is_eman1_helicion(boxpath)
                if is_helicion_ptcl or is_helicion_eman:
                    grid_w = grid_dims[1]
                    num_patches = grid_dims[2]
                    cell_h = img_height / (num_patches * grid_w)

                    if is_helicion_ptcl:
                        filaments = CoordsIO.read_eman1_helicon(
                            boxpath, int(cell_h)
                        )
                    else:
                        filaments = CoordsIO.read_eman1_filament_start_end(boxpath, int(cell_h))

                    for filament in filaments:
                        for box in filament.boxes:
                            obj = {}
                            box_xmin = int(box.x)
                            box_width = int(box.w)
                            box_height = int(box.h)
                            box_ymin = img_height - (int(box.y) + box_height)

                            box_xmax = box_xmin + box_width
                            box_ymax = box_ymin + box_height
                            obj["name"] = "particle"
                            obj["xmin"] = box_xmin
                            obj["ymin"] = box_ymin
                            obj["xmax"] = box_xmax
                            obj["ymax"] = box_ymax
                            img["object"] += [obj]
                            if obj["name"] in seen_labels:
                                seen_labels[obj["name"]] += 1
                            else:
                                seen_labels[obj["name"]] = 1
                elif ann.endswith((".star")):
                    if os.stat(boxpath).st_size != 0:
                        box_lines = np.atleast_2d(
                            np.genfromtxt(boxpath, comments="_", skip_header=4)
                        )
                        for row in box_lines:
                            obj = {}
                            box_xmin = int(row[0] - anchor_size / 2)
                            box_width = int(anchor_size)
                            box_height = int(anchor_size)
                            # box_ymin = img_height - (int(row[1]) + box_height)

                            box_ymin = img_height - int(row[1] + anchor_size / 2)

                            box_xmax = box_xmin + box_width
                            box_ymax = box_ymin + box_height
                            obj["name"] = "particle"
                            obj["xmin"] = box_xmin
                            obj["ymin"] = box_ymin
                            obj["xmax"] = box_xmax
                            obj["ymax"] = box_ymax
                            img["object"] += [obj]
                            if obj["name"] in seen_labels:
                                seen_labels[obj["name"]] += 1
                            else:
                                seen_labels[obj["name"]] = 1
                else:
                    if os.stat(boxpath).st_size != 0:
                        boxes = CoordsIO.read_eman1_boxfile(boxpath)
                        for box in boxes:
                            """
                            Box files are written with coordinate system with an origin in the top left corner. 
                            Each box file is specified by the lower left corner of the box and witdh and the hight of the box.
                            This has to be converted to a coordinate system with origin in the botten left corner.
                            """
                            obj = {}
                            box_xmin = int(box.x)
                            box_width = int(box.w)
                            box_height = int(box.h)
                            box_ymin = img_height - (int(box.y) + box_height)

                            box_xmax = box_xmin + box_width
                            box_ymax = box_ymin + box_height
                            obj["name"] = "particle"
                            obj["xmin"] = box_xmin
                            obj["ymin"] = box_ymin
                            obj["xmax"] = box_xmax
                            obj["ymax"] = box_ymax
                            img["object"] += [obj]
                            if obj["name"] in seen_labels:
                                seen_labels[obj["name"]] += 1
                            else:
                                seen_labels[obj["name"]] = 1

                if len(img["object"]) >= 0:
                    all_imgs += [img]

    return all_imgs, seen_labels


def is_helicon_with_particle_coords(path):
    try:
        with open(path) as f:
            first_line = f.readline()
            f.close()
        return "#micrograph" in first_line
    except ValueError:
        return False


def is_eman1_helicion(path):
    try:
        if os.stat(path).st_size == 0:
            return False
        box_lines = np.atleast_2d(np.genfromtxt(path))
        return (
            len(box_lines) > 1
            and len(box_lines[0]) == 5
            and box_lines[0][4] == -1
            and box_lines[1][4] == -2
        )
    except ValueError:
        return False



#
# COPYRIGHT
#
# All contributions by Ngoc Anh Huyn:
# Copyright (c) 2017, Ngoc Anh Huyn.
# All rights reserved.
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
import numpy as np
import sys

from . import imagereader

class Filament:
    def __init__(self, boxes=None):

        if boxes is None:
            self.boxes = []
        else:
            self.boxes = boxes

    def add_box(self, box):
        self.boxes.append(box)
        return self

    def get_num_boxes(self):
        return len(self.boxes)

    def does_overlap(self, box, treshold):

        for filbox in self.boxes:
            if bbox_iou(filbox, box) >= treshold:
                return True
        return False


class BoundBox:
    def __init__(self, x, y, w, h, c=None, classes=None):
        self.x = x
        self.y = y
        self.w = w
        self.h = h

        self.c = c
        self.classes = classes

        self.label = -1
        self.score = -1
        self.info = None  # helping data during processing

    def get_label(self):
        if self.label == -1:
            self.label = np.argmax(self.classes)

        return self.label

    def get_score(self):
        # if self.score == -1:
        self.score = self.classes[self.get_label()]

        return self.score


class WeightReader:
    def __init__(self, weight_file):
        self.offset = 4
        self.all_weights = np.fromfile(weight_file, dtype="float32")

    def read_bytes(self, size):
        self.offset = self.offset + size
        return self.all_weights[self.offset - size : self.offset]

    def reset(self):
        self.offset = 4


def getEquidistantBoxes(box1, box2, parts):
    points = zip(
        np.linspace(box1.x, box2.x, parts + 1, endpoint=True),
        np.linspace(box1.y, box2.y, parts + 1, endpoint=True),
    )
    new_boxes = []
    if box1.c is None or box2.c is None:
        c = 1
    else:
        c = (box1.c + box2.c) / 2
    classes = ""
    if box1.classes:
        classes = box1.classes
    w = box1.w
    h = box2.w
    for point in points:
        b = BoundBox(x=point[0], y=point[1], w=w, h=h, c=c, classes=classes)
        new_boxes.append(b)
    return new_boxes

def resample_filaments(new_filaments, box_distance):
    resamples_filaments = []
    for fil in new_filaments:
        res_fil = resample_filament(fil, box_distance)
        if not res_fil.boxes:
            prevbox = None
            for box in fil.boxes:
                if prevbox is None:
                    print(box.x, box.y)
                else:
                    print(
                        box.x,
                        box.y,
                        np.sqrt(box_squared_distance(prevbox, box)),
                    )
                prevbox = box
        else:
            resamples_filaments.append(res_fil)
    return resamples_filaments

def resample_filament(filament, distance):

    # Find two boxes with maximum distance
    boxes = filament.boxes
    max_distance = 0
    max_boxa = None
    max_boxb = None

    for i in range(len(boxes)):
        for j in range(len(boxes)):
            if i == j:
                continue
            dist = box_squared_distance(boxes[i], boxes[j])
            if dist > max_distance:
                max_boxa = i
                max_boxb = j
    if max_boxa is not None:
        # Construct a polyline, start from max_boxa
        polyline = []
        polyline.append(boxes[max_boxa])
        assigned_boxes = []
        assigned_boxes.append(max_boxa)
        last_appended_box = boxes[max_boxa]
        last_appended_index = max_boxa
        added_new = True

        while added_new:
            min_distance = 99999999
            nearest_box_index = None
            added_new = False
            for i in range(len(boxes)):
                if i == last_appended_index:
                    continue
                if i in assigned_boxes:
                    continue

                measured_distance = box_squared_distance(last_appended_box, boxes[i])

                if measured_distance < min_distance:
                    nearest_box_index = i
                    min_distance = measured_distance

            if nearest_box_index is not None:
                polyline.append(boxes[nearest_box_index])
                assigned_boxes.append(nearest_box_index)
                last_appended_box = boxes[nearest_box_index]
                last_appended_index = nearest_box_index
                added_new = True

        # Create new boxes based on the polyline
        new_boxes_candidates = []
        sqdistance = distance * distance

        for i in range(len(polyline) - 1):
            eq_boxes = getEquidistantBoxes(polyline[i], polyline[i + 1], parts=100)
            new_boxes_candidates.extend(eq_boxes)
        new_boxes = []
        dist = -1
        for box in new_boxes_candidates:
            if dist == -1:
                new_boxes.append(box)
                dist = 0
            else:
                dist = box_squared_distance(new_boxes[len(new_boxes) - 1], box)
                if dist > sqdistance:
                    new_boxes.append(box)

        new_filament = Filament()
        for box in new_boxes:
            new_filament.add_box(box)

        return new_filament
    return None


def box_squared_distance(boxa, boxb):
    return np.square(boxa.x - boxb.x) + np.square(boxa.y - boxb.y)


def box_to_N_x_M_array(boxes):
    data = np.empty(shape=(len(boxes), 2))

    for i in range(len(boxes)):
        data[i][0] = boxes[i].x
        data[i][1] = boxes[i].y
    return data


def normalize(image, margin_size=0):
    if not np.issubdtype(image.dtype, np.float32):
        image = image.astype(np.float32)
    mask = np.s_[ int(image.shape[0]*(margin_size)):int(image.shape[0]*(1-margin_size)), int(image.shape[1]*margin_size):int(image.shape[1]*(1-margin_size))]
    img_mean = np.mean(image[mask])
    img_std = np.std(image[mask])

    image = (image - img_mean) / (3 * 2 * img_std)


    return image


def bbox_iou(box1, box2):

    x1_min = box1.x - box1.w / 2
    x1_max = box1.x + box1.w / 2
    y1_min = box1.y - box1.h / 2
    y1_max = box1.y + box1.h / 2

    x2_min = box2.x - box2.w / 2
    x2_max = box2.x + box2.w / 2
    y2_min = box2.y - box2.h / 2
    y2_max = box2.y + box2.h / 2

    intersect_w = interval_overlap([x1_min, x1_max], [x2_min, x2_max])
    intersect_h = interval_overlap([y1_min, y1_max], [y2_min, y2_max])

    intersect = intersect_w * intersect_h

    union = box1.w * box1.h + box2.w * box2.h - intersect

    return float(intersect) / union


def interval_overlap(interval_a, interval_b):
    x1, x2 = interval_a
    x3, x4 = interval_b

    if x3 < x1:
        if x4 < x1:
            return 0
        else:
            return min(x2, x4) - x1
    else:
        if x2 < x3:
            return 0
        else:
            return min(x2, x4) - x3


def decode_netout(netout, obj_threshold, nms_threshold, anchors, nb_class):
    grid_h, grid_w, nb_box = netout.shape[:3]

    boxes = []

    # decode the output by the network
    netout[..., 4] = sigmoid(netout[..., 4])
    netout[..., 5:] = netout[..., 4][..., np.newaxis] * softmax(netout[..., 5:])
    netout[..., 5:] *= netout[..., 5:] > obj_threshold

    for row in range(grid_h):
        for col in range(grid_w):
            for b in range(nb_box):
                # from 4th element onwards are confidence and class classes
                classes = netout[row, col, b, 5:]

                if np.sum(classes) > 0:
                    # first 4 elements are x, y, w, and h
                    x, y, w, h = netout[row, col, b, :4]

                    x = (
                        col + sigmoid(x)
                    ) / grid_w  # center position, unit: image width
                    y = (
                        row + sigmoid(y)
                    ) / grid_h  # center position, unit: image height
                    w = anchors[2 * b + 0] * np.exp(w) / grid_w  # unit: image width
                    h = anchors[2 * b + 1] * np.exp(h) / grid_h  # unit: image height
                    confidence = netout[row, col, b, 4]

                    box = BoundBox(x, y, w, h, confidence, classes)

                    boxes.append(box)

    # suppress non-maximal boxes
    for c in range(nb_class):
        sorted_indices = list(reversed(np.argsort([box.classes[c] for box in boxes])))

        for i in range(len(sorted_indices)):
            index_i = sorted_indices[i]

            if boxes[index_i].classes[c] == 0:
                continue
            else:
                for j in range(i + 1, len(sorted_indices)):
                    index_j = sorted_indices[j]

                    if bbox_iou(boxes[index_i], boxes[index_j]) >= nms_threshold:
                        boxes[index_j].classes[c] = 0

    # remove the boxes which are less likely than a obj_threshold
    boxes = [box for box in boxes if box.get_score() > obj_threshold]

    return boxes


def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))


def softmax(x, axis=-1, t=-100.0):
    x = x - np.max(x)

    if np.min(x) < t:
        x = x / np.min(x) * t

    e_x = np.exp(x)

    return e_x / e_x.sum(axis, keepdims=True)


def filter_images_nn_img(img_path, model_path, padding=15, batch_size=4):
    import h5py
    from janni import predict as janni_predict
    from janni import models as janni_models
    with h5py.File(model_path, mode="r") as f:
        try:
            import numpy as np
            model = str(np.array((f["model_name"])))
            patch_size = tuple(f["patch_size"])
        except KeyError:
            print("Not supported filtering model. Stop.")
            sys.exit(0)

    if model == "unet":
        model = janni_models.get_model_unet(input_size=patch_size)
        model.load_weights(model_path)
    else:
        print("Not supported model", model, "Stop")
        sys.exit(0)

    image = imagereader.image_read(img_path)
    fitlered_img = janni_predict.predict_np(model,
                                              image,
                                              patch_size=(1024, 1024),
                                              padding=padding,
                                              batch_size=batch_size)

    return fitlered_img

def filter_images_nn_dir(img_paths, output_dir_filtered_imgs, model_path, padding=15, batch_size=4):
    import h5py
    from janni import predict as janni_predict
    from janni import models as janni_models
    with h5py.File(model_path, mode="r") as f:
        try:
            import numpy as np
            model = str(np.array((f["model_name"])))
            patch_size = tuple(f["patch_size"])
        except KeyError:
            print("Not supported filtering model. Stop.")
            sys.exit(0)

    if model == "unet":
        model = janni_models.get_model_unet(input_size=patch_size)
        model.load_weights(model_path)
    else:
        print("Not supported model", model, "Stop")
        sys.exit(0)


    filtered_paths = janni_predict.predict_list(
        image_paths = img_paths,
        output_path = output_dir_filtered_imgs,
        model=model,
        patch_size=patch_size,
        padding=padding,
        batch_size=batch_size)

    return filtered_paths


"""
Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
"""
#
# COPYRIGHT
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

import random
import numpy as np
from scipy import ndimage


class Augmentation:
    """
    Class for doing data augmentation
    """

    def __init__(self, is_grey=False):
        self.is_grey = is_grey

    def image_augmentation(self, image):
        """
        Applies random selection of data augmentations
        :param image:  Input image
        :return: Augmented image
        """
        augmentations = [
            self.additive_gauss_noise,
            self.add,
            self.contrast_normalization,
            self.multiply,
           # self.sharpen,
            self.dropout,
        ]

        num_augs = np.random.randint(0, np.minimum(6, len(augmentations)))
        if num_augs > 0:

            if np.random.rand() > 0.5:
                augmentations.append(self.gauss_blur)
            else:
                augmentations.append(self.avg_blur)

            selected_augs = random.sample(augmentations, num_augs)
            image = image.astype(np.float32, copy=False)
            for sel_aug in selected_augs:
                image = sel_aug(image)
            #   print "Mean after", sel_aug, " sum: ", np.mean(image)
            if self.is_grey:
                min_img = np.min(image)
                max_img = np.max(image)
                image = ((image - min_img) / (max_img - min_img)) * 255
                #    image = np.clip(image, 0, 255)
                image = image.astype(np.uint8, copy=False)

        return image

    def gauss_blur(self, image, sigma_range=(0, 3)):
        """
        Applys gaussian blurring with random sigma
        :param image: Input image
        :param sigma_range:  Range for random sigma
        :return: Blurred image
        """
        rand_sigma = sigma_range[0] + np.random.rand() * (
            sigma_range[1] - sigma_range[0]
        )
        result = ndimage.gaussian_filter(image, sigma=rand_sigma, mode="nearest", output=image)

        if not np.issubdtype(image.dtype, np.float32):
            result = result.astype(np.float32, copy=False)
        return result

    def avg_blur(self, image, kernel_size=(2, 7)):
        """
        Applys average blurring with random kernel size
        :param image: Input image (numpy array)
        :param kernel_size: Range for random kernel size
        :return: Blurred image
        """
        rang_kernel_size = np.random.randint(kernel_size[0], kernel_size[1])
        image = ndimage.uniform_filter(image, size=rang_kernel_size, mode="nearest", output=image)
        return image


    def additive_gauss_noise(self, image, max_sigma_range_factor=0.05):
        """
        Add random gaussian noise to image
        :param image: Input image
        :param max_sigma_range_factor: Range for max_sigma_range. The standard deviation of the noise is
        choosen randomly depending on the standard deviation of the image.
        The choosen standard deviation for noise is between: 0 and  max_sigma_factor*6*np.std(image)
        :return:
        """

        width = 2 * 3 * np.std(image)
        max_sigma = width * max_sigma_range_factor
        rand_sigma = np.random.rand() * max_sigma
        noise = np.random.randn(image.shape[0], image.shape[1])

        #image = noise*rand_sigma + image

        np.multiply(noise, rand_sigma, out=noise)
        np.add(image, noise, out=image)

        if not np.issubdtype(image.dtype, np.float32):
            image = image.astype(np.float32, copy=False)

        return image

    def contrast_normalization(self, image, alpha_range=(0.5, 2.0)):
        """
        Spread or squeeze the pixel values.
        :param image: Input image
        :param alpha_range: Range for alpha. Alpha controlls the normalization.
        :return: Modified image
        """
        rand_multiplyer = alpha_range[0] + np.random.rand() * (
            alpha_range[1] - alpha_range[0]
        )
        middle = np.median(image)
        np.subtract(image, middle, out=image)
        np.multiply(rand_multiplyer, image, out=image)
        np.add(middle, image, out=image)

        return image

    def dropout(self, image, ratio=(0.01, 0.1)):
        """
        Set a random selection of pixels to the mean of the image
        :param image: Input image
        :param ratio: Range for random ratio
        :return: Modified image
        """
        if isinstance(ratio, float):
            rand_ratio = ratio
        else:
            rand_ratio = ratio[0] + np.random.rand() * (ratio[1] - ratio[0])
        mean_val = np.mean(image)
        drop = np.random.binomial(
            n=1, p=1 - rand_ratio, size=(image.shape[0], image.shape[1])
        )
        image[drop == 0] = mean_val

        return image

    def add(self, image, scale=0.05):
        """
        Adds a random constant to the image
        :param image: Input image
        :param scale: Scale for random constant. The random constant
        will be between 0 and scale*6*std(image)
        :return: Modified image
        """
        width = 2 * 3 * np.std(image)
        width_rand = scale * width
        rand_constant = (np.random.rand() * width_rand) - width_rand / 2
        np.add(image, rand_constant, out=image)

        return image

    def multiply(self, image, mult_range=(0.5, 1.5)):
        """
        Multipy the input image by a random float.
        :param image: Input image
        :param mult_range: Range for random multiplier
        :return: multiplied image
        """

        rand_multiplyer = mult_range[0] + np.random.rand() * (
            mult_range[1] - mult_range[0]
        )
        np.multiply(image, rand_multiplyer, out=image)
        return image



#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

import sys


def generate_config_file(config_out_path,
                         architecture,
                         input_size,
                         anchors,
                         max_box_per_image,
                         num_patches,
                         overlap_patches,
                         filter,
                         train_image_folder,
                         train_annot_folder,
                         train_times,
                         pretrained_weights,
                         batch_size,
                         learning_rate,
                         nb_epoch,
                         object_scale,
                         no_object_scale,
                         coord_scale,
                         class_scale,
                         log_path,
                         saved_weights_name,
                         debug,
                         valid_image_folder,
                         valid_annot_folder,
                         valid_times):

    model_dict = {'architecture': architecture,
                  'input_size': input_size,
                  'anchors': anchors,
                  'overlap_patches': overlap_patches,
                  'num_patches': num_patches,
                  'max_box_per_image': max_box_per_image,
                  }

    train_dict = {'train_image_folder': train_image_folder,
                  'train_annot_folder': train_annot_folder,
                  'train_times': train_times,
                  'pretrained_weights': pretrained_weights,
                  'batch_size': batch_size,
                  'learning_rate': learning_rate,
                  'nb_epoch': nb_epoch,
                  'object_scale': object_scale,
                  'no_object_scale': no_object_scale,
                  'coord_scale': coord_scale,
                  'class_scale': class_scale,
                  "saved_weights_name": saved_weights_name,
                  "debug": debug,
                  "log_path": log_path
                  }

    valid_dict = {'valid_image_folder': valid_image_folder,
                  'valid_annot_folder': valid_annot_folder,
                  'valid_times': valid_times
                  }

    if filter is not None:
        model_dict["filter"] = filter

    from json import dump
    dict = {"model": model_dict, "train": train_dict, "valid": valid_dict}
    with open(config_out_path, 'w') as f:
        dump(dict, f, ensure_ascii=False, indent=4)
    print("Wrote config to",config_out_path)


def get_number_patches(config):
    """
    Returns the number of patches in config file.
    :param config: config dictionary
    :return: Number of patches
    """
    if "num_patches" in config["model"]:
        num_patches = config["model"]["num_patches"]
    else:
        num_patches = 1
        print("Set patch_size to 1 (default)")

    return num_patches


def get_gridcell_dimensions(config):
    """
    Returns the grid cell dimension in dependence of the use network.
    :param config: config dictionary
    :return: grid cell dimensions
    """
    if config["model"]["architecture"] == "YOLO":
        downsampling_factor = 32.0
    elif (
        config["model"]["architecture"] == "crYOLO"
        or config["model"]["architecture"] == "PhosaurusNet"
    ):
        downsampling_factor = 16.0
    else:
        raise Exception(
            "Architecture not supported! "
            "Only support for PhosaurusNet, YOLO and crYOLO at the moment!"
        )

    grid_w = config["model"]["input_size"] / downsampling_factor
    grid_h = config["model"]["input_size"] / downsampling_factor

    return grid_w, grid_h


def get_box_size(config):
    """
    Read the box size from the config.

    :param config:
    :return: Box size
    """
    if "particle_diameter" in config["model"]:
        box_height = config["model"]["particle_diameter"]
        box_width = config["model"]["particle_diameter"]
    elif len(config["model"]["anchors"]) == 2:
        box_height, box_width = config["model"]["anchors"]
    else:
        sys.exit("You have to specify the 'particle_diameter' in your config. Stop.")
    return box_width, box_height



"""
Prediction script crYOLO
"""

# ! /usr/bin/env python
#
# COPYRIGHT
#
# All contributions by Ngoc Anh Huyn:
# Copyright (c) 2017, Ngoc Anh Huyn.
# All rights reserved.
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
from __future__ import print_function
import multiprocessing
import time
import argparse
import os
import sys
import json
import traceback
import numpy as np
from lineenhancer import line_enhancer, maskstackcreator
from . import CoordsIO
from . import imagereader
from . import utils
from . import filament_tracer
from . import config_tools
from . import lowpass
from gooey import Gooey, GooeyParser

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
try:
    os.environ["CUDA_VISIBLE_DEVICES"]
except KeyError:
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"


filament_tracers = None
ARGPARSER = None


def create_parser(parser, use_gooed=True):

    required_group = parser.add_argument_group(
        "Required arguments", "These options are mandatory to run crYOLO prediction"
    )
    args = ["-c", "--conf"]
    kwargs = {
        "help": "Path to configuration file.",
        "required": True,
        "gooey_options": {
            "validator": {
                "test": 'user_input.endswith("json")',
                "message": "File has to end with .json!",
            },
            "wildcard": "*.json"
        },
    }

    if use_gooed:
        kwargs["widget"] = "FileChooser"
    required_group.add_argument(*args, **kwargs)

    args = ["-w", "--weights"]
    kwargs = {
        "help": "Path to the trained model (.h5 file).",
        "required": True,
        "gooey_options": {
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            },
            "wildcard": "*.h5"
        },
    }

    if use_gooed:
        kwargs["widget"] = "FileChooser"
    required_group.add_argument(*args, **kwargs)

    args = ["-i", "--input"]
    kwargs = {
        "nargs": "+",
        "help": "Path to one or multiple image folders / images.",
        "required": True,
    }
    if use_gooed:
        kwargs["widget"] = "DirChooser"
    required_group.add_argument(*args, **kwargs)

    args = ["-o", "--output"]
    kwargs = {"help": "Specify the output folder.", "required": True}
    if use_gooed:
        kwargs["widget"] = "DirChooser"

    required_group.add_argument(*args, **kwargs)

    optional_group = parser.add_argument_group(
        "Optional arguments", "Optional arguments for crYOLO"
    )

    optional_group.add_argument(
        "-t",
        "--threshold",
        type=float,
        default=0.3,
        help="Confidence threshold. Have to be between 0 and 1. The higher, the more conservative.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0.0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "-g",
        "--gpu",
        default="",
        # type=int,
        nargs="+",
        help="Specifiy which gpu(s) should be used. Multiple GPUs are separated by a whitespace.",
    )

    optional_group.add_argument(
        "-d",
        "--distance",
        default=0,
        type=int,
        help="Particles with a distance less than this value (in pixel) will be removed",
    )

    optional_group.add_argument(
        "-pbs",
        "--prediction_batch_size",
        default=3,
        type=int,
        help="How many images should be predicted in one batch. Smaller values might resolve memory issues.",
    )

    optional_group.add_argument(
        "--gpu_fraction",
        type=float,
        default=1.0,
        help="Specify the fraction of memory per GPU used by crYOLO during prediction. Only values between 0.0 and 1.0 are allowed.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0.0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "-nc",
        "--num_cpu",
        type=int,
        default=multiprocessing.cpu_count(),
        help="Number of CPUs used during filtering / filament tracing. By default it will use all of the available CPUs.",
    )

    optional_group.add_argument(
        "--norm_margin",
        type=float,
        default=0.0,
        help="Relative margin size for normalization.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0.0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "--otf",
        action="store_true",
        default=False,
        help="On the fly filtering. Filtered micrographs will not be written to disk. Might be slower",
    )



    filament_group = parser.add_argument_group(
        "Filament options",
        "These options are only relevant if you want to use  the filament mode",
    )

    filament_group.add_argument(
        "--filament", action="store_true", help="Activate filament mode"
    )

    filament_group.add_argument(
        "--nosplit",
        action="store_true",
        help="The filament mode does not split to curved filaments",
    )

    filament_group.add_argument(
        "--nomerging",
        action="store_true",
        help="The filament mode does not merge filaments",
    )

    filament_group.add_argument(
        "-fw",
        "--filament_width",
        default=None,
        type=int,
        help="Filament width (in pixel)",
    )

    filament_group.add_argument(
        "-mw", "--mask_width", default=100, type=int, help="Mask width (in pixel)"
    )

    filament_group.add_argument(
        "-bd",
        "--box_distance",
        default=None,
        type=int,
        help="Distance between two boxes(in pixel)",
    )

    filament_group.add_argument(
        "-mn",
        "--minimum_number_boxes",
        default=None,
        type=int,
        help="Minimum number of boxes per filament.",
    )

    filament_group.add_argument(
        "-sr",
        "--search_range_factor",
        type=float,
        default=1.41,
        help="The search range for connecting boxes is the box size times this factor.",
    )

    depexp_group = parser.add_argument_group(
        "Deprecated/Experimental/Special ",
        "Contains either deprecated / experimental or very special options.",
    )
    depexp_group.add_argument(
        "-p",
        "--patch",
        type=int,
        help="(DEPRECATED) Number of patches. Use the config file in case you want to use patches.",
    )

    depexp_group.add_argument(
        "--write_empty",
        action="store_true",
        help="Write empty box files when not particle could be found.",
    )


def get_parser():
    parser = GooeyParser(
        description="Pick particles with crYOLO on any dataset",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    create_parser(parser)
    return parser


def _main_():
    import sys

    # if sys.argv[1] == "gui":
    if len(sys.argv) >= 2:
        if not "--ignore-gooey" in sys.argv:
            sys.argv.append("--ignore-gooey")

    # r'^\d+ particles are found in .* \( (\d+) % \)$'
    kwargs = {"terminal_font_family": "monospace",
              "richtext_controls": True}
    Gooey(
        main,
        program_name="crYOLO Predict",
        image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        **kwargs
    )()


def main(args=None):
    if args is None:
        parser = get_parser()
        args = parser.parse_args()

    if isinstance(args.gpu, list):
        if len(args.gpu) == 1:
            num_gpus = 1
            if args.gpu[0] != "-1":
                str_gpus = args.gpu[0].strip().split(" ")
                num_gpus = len(str_gpus)
                os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
        else:
            str_gpus = [str(entry) for entry in args.gpu]
            num_gpus = len(str_gpus)
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
    else:
        num_gpus = 1
        if args.gpu != -1 and len(args.gpu) > 0:
            str_gpus = str(args.gpu)
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)

    if args.gpu_fraction < 1.0 and args.gpu_fraction > 0.0:
        import tensorflow as tf
        from keras.backend.tensorflow_backend import set_session

        config = tf.ConfigProto()
        config.gpu_options.per_process_gpu_memory_fraction = args.gpu_fraction
        set_session(tf.Session(config=config))

    config_path = args.conf
    weights_path = args.weights
    input_path = args.input
    obj_threshold = args.threshold
    min_distance = args.distance
    mask_width = args.mask_width
    prediction_batch_size = args.prediction_batch_size
    no_merging = args.nomerging
    otf = args.otf
    do_merging = True
    norm_margin = args.norm_margin
    search_range_factor = args.search_range_factor
    if no_merging:
        do_merging = False

    nosplit = False
    if args.nosplit is not None:
        nosplit = args.nosplit

    outdir = None
    if args.output is not None:
        outdir = str(args.output)
    write_empty = args.write_empty

    num_cpus = multiprocessing.cpu_count()
    if args.num_cpu != -1:
        num_cpus = args.num_cpu

    with open(config_path) as config_buffer:
        try:
            config = json.load(config_buffer)
        except json.JSONDecodeError:
            print(
                "Your configuration file seems to be corruped. Please check if it is valid."
            )
        if config["model"]["input_size"] % 32 > 0:
            new_input_size = int(32 * round(float(config["model"]["input_size"]) / 32))
            print(
                "You input size is not a multiple of 32. Round it to the next multiple of 32:",
                new_input_size,
            )
            config["model"]["input_size"] = new_input_size

    filament_mode = args.filament
    if filament_mode:
        if args.filament_width is None:
            sys.exit("Please specifiy your filament width ( -fw / --filament_width)")
        else:
            filament_width = args.filament_width
        if args.box_distance is None:
            sys.exit("Please specifiy your box distance ( -bd / --box_distance)")
        else:
            box_distance = args.box_distance
        minimum_number_boxes = 1
        if args.minimum_number_boxes is not None:
            minimum_number_boxes = args.minimum_number_boxes

    if args.patch is not None and args.patch > 0:
        num_patches = int(args.patch)
    else:
        num_patches = config_tools.get_number_patches(config)

    # Get overlap patches
    overlap_patches = 0
    if "overlap_patches" in config["model"]:
        overlap_patches = int(config["model"]["overlap_patches"])
    elif not len(config["model"]["anchors"]) > 2:
        overlap_patches = config["model"]["anchors"][0]

    picking_results = do_prediction(
        config_path=config_path,
        weights_path=weights_path,
        input_path=input_path,
        obj_threshold=obj_threshold,
        num_patches=num_patches,
        filament_mode=filament_mode,
        write_empty=write_empty,
        overlap=overlap_patches,
        num_images_batch_prediction=prediction_batch_size,
        num_gpus=num_gpus,
        num_cpus=num_cpus,
        otf=otf,
        normalization_margin=norm_margin,
    )

    # Resacle to image dimensions, resize boxes

    if "particle_diameter" in config["model"]:
        if isinstance(config["model"]["particle_diameter"], list):
            export_size = config["model"]["particle_diameter"][0]
        else:
            export_size = config["model"]["particle_diameter"]
    elif len(config["model"]["anchors"]) == 2:
        export_size = config["model"]["anchors"][0]
    else:
        export_size = None

    # Rescaling

    for picking_result_micrograph in picking_results:

        # if num_patches > 1:
        #    image_width = int(picking_result_micrograph["img_width"]/num_patches)+2*overlap_patches
        #    image_height = int(picking_result_micrograph["img_height"]/num_patches)+2*overlap_patches
        # else:
        image_width = picking_result_micrograph["img_width"]
        image_height = picking_result_micrograph["img_height"]
        picking_result_micrograph["boxes"] = [
            rescale(box, image_height, image_width, export_size)
            for box in picking_result_micrograph["boxes"]
        ]

        picking_result_micrograph["boxes_unfiltered"] = [
            rescale(box, image_height, image_width, export_size)
            for box in picking_result_micrograph["boxes_unfiltered"]
        ]

    # Min distance filter
    if min_distance > 0:
        sum_particles_before_filtering = np.sum(
            [len(res["boxes"]) for res in picking_results]
        )
        for picking_result_micrograph in picking_results:
            picking_result_micrograph["boxes"] = min_distance_filter(
                picking_result_micrograph["boxes"], min_distance
            )

        num_remain_particles = np.sum([len(res["boxes"]) for res in picking_results])
        print(
            sum_particles_before_filtering - num_remain_particles,
            "particles were filtered because of the distance threshold",
            min_distance,
        )

        print("Total picked particles after distance filtering:", num_remain_particles)

    # Filtering of particles which are not fully immersed in the micrograph
    deleted = 0
    for picking_result_micrograph in picking_results:
        image_width = picking_result_micrograph["img_width"]
        image_height = picking_result_micrograph["img_height"]

        boxes = picking_result_micrograph["boxes"]
        boxes_to_delete = get_not_fully_immersed_box_indices(
            boxes, image_height, image_width
        )
        for index in sorted(boxes_to_delete, reverse=True):
            del boxes[index]
        deleted += len(boxes_to_delete)

        boxes = picking_result_micrograph["boxes_unfiltered"]
        boxes_to_delete = get_not_fully_immersed_box_indices(
            boxes, image_height, image_width
        )
        for index in sorted(boxes_to_delete, reverse=True):
            del boxes[index]
    print(
        "Deleted",
        deleted,
        "particles as they were not fully immersed in the micrograph",
    )

    ###############################
    #   Filament Post Processing
    ###############################

    # 1. Build sets of images (size = number of processors)
    # 2. Enhance
    # 3. Filamental post processing

    if filament_mode:
        picking_result_with_boxes = []
        picking_result_no_boxes = []
        picking_result_with_boxes_subsets = []
        picked_filaments = 0
        for picking_result_micrograph in picking_results:
            if picking_result_micrograph["boxes"]:
                picking_result_with_boxes.append(picking_result_micrograph)
            else:
                picking_result_micrograph["filaments"] = []
                picking_result_no_boxes.append(picking_result_micrograph)

        if picking_result_with_boxes:
            image_width, image_height = imagereader.read_width_height(
                picking_result_with_boxes[0]["img_path"]
            )
            rescale_factor = 1024.0 / max(image_width, image_height)
            rescale_factor_x = 1024.0 / image_width
            rescale_factor_y = 1024.0 / image_height

            mask_creator = maskstackcreator.MaskStackCreator(
                filament_width=filament_width * rescale_factor,
                mask_size=1024,
                mask_width=mask_width,
                angle_step=2,
                bright_background=True,
            )
            print("Start filament tracing")
            print("Initialisation mask stack")
            mask_creator.init()
            # Devide picking result into chunks

            number_processors = num_cpus

            picking_result_with_boxes_subsets = [
                picking_result_with_boxes[i : i + number_processors]
                for i in range(0, len(picking_result_with_boxes), number_processors)
            ]
            process_counter = 1

            # Parallel tracing

            filament_width_scaled = filament_width * rescale_factor
            search_radius_scaled = export_size * rescale_factor * search_range_factor
            for picking_result_subset in picking_result_with_boxes_subsets:
                image_subset = [
                    picking_result_subset[i]["img_path"]
                    for i in range(0, len(picking_result_subset))
                ]
                boxes_subset = [
                    picking_result_subset[i]["boxes"]
                    for i in range(0, len(picking_result_subset))
                ]
                print(
                    " Enhance subset ",
                    process_counter,
                    "of",
                    len(picking_result_with_boxes_subsets),
                )
                enhanced_images = line_enhancer.enhance_images(
                    image_subset, mask_creator
                )
                angle_images = [
                    enhanced_images[i]["max_angle"] for i in range(len(enhanced_images))
                ]

                print(
                    " Trace subset ",
                    process_counter,
                    "of",
                    len(picking_result_with_boxes_subsets),
                )
                global filament_tracers
                filament_tracers = []
                for index, boxset in enumerate(boxes_subset):
                    angle_image_flipped = np.flipud(angle_images[index])
                    tracer = filament_tracer.FilamentTracer(
                        boxes=boxset,
                        orientation_image=angle_image_flipped,
                        filament_width=filament_width_scaled,
                        search_radius=search_radius_scaled,
                        angle_delta=10,
                        rescale_factor=rescale_factor,
                        rescale_factor_x=rescale_factor_x,
                        rescale_factor_y=rescale_factor_y,
                        do_merging=do_merging,
                        box_distance=box_distance,
                    )
                    filament_tracers.append(tracer)

                pool = multiprocessing.Pool(processes=num_cpus)

                subset_new_filaments = pool.map(
                    trace_subset_filements, range(len(filament_tracers)), chunksize=1
                )
                pool.close()
                pool.join()

                for index_subset in range(len(picking_result_subset)):
                    new_filaments = subset_new_filaments[index_subset]

                    resamples_filaments = utils.resample_filaments(
                        new_filaments, box_distance
                    )

                    # Make straight filaments
                    if not nosplit:
                        resamples_filaments = filament_tracer.split_filaments_by_straightness(
                            resamples_filaments
                        )

                    # NMS for filaments:
                    resamples_filaments = filament_tracer.nms_for_filaments(
                        resamples_filaments, filament_width
                    )

                    # Min number of boxes filter:
                    resamples_filaments = filament_tracer.filter_filaments_by_num_boxes(
                        resamples_filaments, minimum_number_boxes
                    )

                    picked_filaments += len(resamples_filaments)

                    if len(resamples_filaments) >= 1:
                        picking_result_subset[index_subset][
                            "filaments"
                        ] = resamples_filaments
                    else:
                        picking_result_subset[index_subset]["filaments"] = None

                print("Total number of filaments picked so far: ", picked_filaments)
                process_counter += 1

        if write_empty:
            picking_result_with_boxes_subsets.append(picking_result_no_boxes)

        print("Total number of filaments picked: ", picked_filaments)

        ###############################
        #   Write bounding boxes
        ###############################
        for picking_result_subset in picking_result_with_boxes_subsets:

            for result in picking_result_subset:

                if (result["filaments"] is not None) or write_empty:
                    pth = result["pth"]
                    eman_helix_segmented_path = pth
                    eman_start_end = pth
                    star_start_end = pth[:-3] + "star"

                    if outdir is not None:
                        filename = os.path.basename(pth)
                        eman_helix_segmented_path = os.path.join(
                            outdir, "EMAN_HELIX_SEGMENTED", filename
                        )
                        eman_start_end = os.path.join(
                            outdir, "EMAN_START_END", filename
                        )
                        filename = filename[:-3] + "star"
                        star_start_end = os.path.join(
                            outdir, "STAR_START_END", filename
                        )

                    if not os.path.exists(os.path.dirname(eman_helix_segmented_path)):
                        os.makedirs(os.path.dirname(eman_helix_segmented_path))

                    if not os.path.exists(os.path.dirname(eman_start_end)):
                        os.makedirs(os.path.dirname(eman_start_end))

                    if not os.path.exists(os.path.dirname(star_start_end)):
                        os.makedirs(os.path.dirname(star_start_end))
                    CoordsIO.write_eman1_helicon(
                        filaments=result["filaments"],
                        path=eman_helix_segmented_path,
                        image_filename=os.path.basename(result["img_path"]),
                    )

                    CoordsIO.write_eman1_filament_start_end(
                        filaments=result["filaments"], path=eman_start_end
                    )

                    CoordsIO.write_star_filemant_file(
                        filaments=result["filaments"], path=star_start_end
                    )

    else:
        ###############################
        #   Write bounding boxes
        ###############################

        for box_to_write in picking_results:

            original_path = box_to_write["pth"]
            eman1_path = original_path

            star_path = os.path.splitext(original_path)[0] + ".star"
            cbox_path = os.path.splitext(original_path)[0] + ".cbox"

            if outdir is not None:
                filename = os.path.basename(eman1_path)
                eman1_path = os.path.join(outdir, "EMAN", filename)

            # Create directory if it does not existes
            if not os.path.exists(os.path.dirname(eman1_path)):
                os.makedirs(os.path.dirname(eman1_path))

            CoordsIO.write_eman1_boxfile(path=eman1_path, boxes=box_to_write["boxes"])

            if outdir is not None:
                filename = os.path.basename(star_path)
                star_path = os.path.join(outdir, "STAR", filename)

            # Create directory if it does not existes
            if not os.path.exists(os.path.dirname(star_path)):
                os.makedirs(os.path.dirname(star_path))
            CoordsIO.write_star_file(path=star_path, boxes=box_to_write["boxes"])

            if outdir is not None:
                filename = os.path.basename(cbox_path)
                star_path = os.path.join(outdir, "CBOX", filename)

            # Create directory if it does not existes
            if not os.path.exists(os.path.dirname(star_path)):
                os.makedirs(os.path.dirname(star_path))
            CoordsIO.write_cbox_file(
                path=star_path, boxes=box_to_write["boxes_unfiltered"]
            )


def min_distance_filter(boxes, min_distance):
    boxes_to_delete = []
    min_distance_sq = min_distance * min_distance
    for box_a_index, box_a in enumerate(boxes):
        for box_b_index in range(box_a_index, len(boxes)):
            if box_a_index != box_b_index:
                distsq = utils.box_squared_distance(box_a, boxes[box_b_index])

                if distsq < min_distance_sq:
                    box_index_to_delte = box_a_index
                    if box_a.c > boxes[box_b_index].c:
                        box_index_to_delte = box_b_index
                    if box_index_to_delte not in boxes_to_delete:
                        boxes_to_delete.append(box_index_to_delte)
    for index in sorted(boxes_to_delete, reverse=True):
        del boxes[index]

    return boxes


def rescale(box, image_height, image_width, export_size=None):
    x_ll = int((box.x - box.w / 2) * image_height)  # lower left
    y_ll = int(
        image_width - box.y * image_width - box.h / 2 * image_width
    )  # lower right
    boxheight_in_pxl = int(box.h * image_width)
    boxwidth_in_pxl = int(box.w * image_height)
    if export_size is not None:
        delta_boxheight = export_size - boxheight_in_pxl
        delta_boxwidth = export_size - boxwidth_in_pxl
        x_ll = x_ll - delta_boxheight / 2
        y_ll = y_ll - delta_boxwidth / 2
        boxheight_in_pxl = export_size
    box.x = x_ll
    box.y = y_ll
    box.w = boxheight_in_pxl
    box.h = boxheight_in_pxl

    return box


def do_prediction(
    config_path,
    weights_path,
    input_path,
    num_patches,
    obj_threshold=0.3,
    write_empty=False,
    config_pre=None,
    overlap=0,
    filament_mode=False,
    num_images_batch_prediction=3,
    num_gpus=1,
    num_cpus=-1,
    yolo=None,
    otf=False,
    normalization_margin=0,
):
    """

    :param config_path: Path to the config file
    :param weights_path: Path do weights file (h5)
    :param input_path: Path to the folder containing the input images
    :param num_patches: Number of patches to use
    :param obj_threshold: Threshold for objects
    :param write_empty:
    :param config_pre:
    :param overlap:
    :param filament_mode:
    :param num_images_batch_prediction:
    :param num_gpus:
    :param yolo:
    :return:
    """
    for path in input_path:
        path_exists = os.path.exists(path)
        if not path_exists:
            sys.exit("Input path does not exist: " + path)

    img_paths = []
    if isinstance(input_path, list):

        for path in input_path:
            isdir = os.path.isdir(path)
            if isdir:
                dir_files = os.listdir(path)

                dir_files = [
                    i
                    for i in dir_files
                    if not i.startswith(".")
                    and os.path.isfile(os.path.join(path, i))
                    and i.endswith(("tiff", "tif", "mrc", "mrcs", "png", "jpg", "jpeg"))
                ]

                img_paths.extend(
                    [os.path.join(path, image_file) for image_file in dir_files]
                )
            elif os.path.isfile(path):
                if not path.startswith(".") and path.endswith(
                    ("tiff", "tif", "mrc", "png", "mrcs", "jpg", "jpeg")
                ):
                    img_paths.append(path)
    else:
        isdir = os.path.isdir(input_path)
        if isdir:
            img_paths = os.listdir(input_path)

            img_paths = [
                i
                for i in img_paths
                if not i.startswith(".")
                and os.path.isfile(os.path.join(input_path, i))
                and i.endswith(("tiff", "tif", "mrc", "mrcs", "png", "jpg", "jpeg"))
            ]

    if not img_paths:
        sys.exit("No valid image in your specified input")
    img_paths.sort()
    first_image_path = img_paths[0]
    if config_pre is not None:
        config = config_pre
    else:
        with open(config_path) as config_buffer:
            config = json.load(config_buffer)

    # Read (first) image and check the image depth.

    try:
        img_first = imagereader.image_read(first_image_path)
    except ValueError:
        sys.exit("Image " + first_image_path + " is not valid")
    if img_first is None:
        sys.exit("No valid image: " + first_image_path)

    if len(img_first.shape) == 2:
        depth = 1
    elif img_first.shape[2] == 1:
        depth = 1
    elif np.all(img_first[:, :, 0] == img_first[:, :, 1]) and np.all(
        img_first[:, :, 0] == img_first[:, :, 2]
    ):
        depth = 1
    else:
        depth = 3

    grid_w, grid_h = config_tools.get_gridcell_dimensions(config)

    #############################################
    # Read meta data about the model
    #############################################
    anchors = None
    yolo_kwargs = {}
    import h5py

    with h5py.File(weights_path, mode="r") as f:
        try:
            anchors = list(f["anchors"])
        except KeyError:
            None

    with h5py.File(weights_path, mode="r") as f:
        try:
            num_free_layers = int(list(f["num_free_layers"])[0])
            yolo_kwargs["num_fine_tune_layers"] = num_free_layers
        except KeyError:
            pass

    if anchors is None:

        if len(config["model"]["anchors"]) > 2:
            # general model is used
            img_width = 4096.0 / num_patches
            img_height = 4096.0 / num_patches
            cell_w = img_width / grid_w
            cell_h = img_height / grid_h
            anchors = np.array(config["model"]["anchors"], dtype=float)
            anchors[::2] = anchors[::2] / cell_w
            anchors[1::2] = anchors[1::2] / cell_h

        else:
            # specifc model is used

            img_width = float(img_first.shape[1]) / num_patches
            img_height = float(img_first.shape[0]) / num_patches
            cell_w = img_width / grid_w
            cell_h = img_height / grid_h

            box_width, box_height = config_tools.get_box_size(config)

            anchor_width = 1.0 * box_width / cell_w
            anchor_height = 1.0 * box_height / cell_h

            anchors = [anchor_width, anchor_height]
        print("Calculated Anchors using first image", anchors)
    else:
        print("Read Anchor from model", anchors)

    if yolo is None:
        ###############################
        #   Make the model
        ###############################
        backend_weights = None
        if "backend_weights" in config["model"]:
            backend_weights = config["model"]["backend_weights"]
        from .frontend import YOLO

        yolo = YOLO(
            architecture=config["model"]["architecture"],
            input_size=config["model"]["input_size"],
            input_depth=depth,
            labels=["particle"],
            max_box_per_image=config["model"]["max_box_per_image"],
            anchors=anchors,
            backend_weights=backend_weights,
            pretrained_weights=weights_path,
            **yolo_kwargs
        )

        ###############################
        #   Load trained weights
        ###############################

        # USE MULTIGPU
        if num_gpus > 1:
            from keras.utils import multi_gpu_model

            parallel_model = multi_gpu_model(yolo.model, gpus=num_gpus)
            yolo.model = parallel_model
    else:
        yolo.anchors = anchors

    ##############################
    # Filter the data
    ##############################

    if otf and not "filter" in config["model"]:
        print(
            "You specified the --otf option. However, filtering is not configured in your"
            "config line, therefore crYOLO will ignore --otf."
        )
    do_nn_filter = False
    if "filter" in config["model"]:

        filter_options = config["model"]["filter"]
        if len(filter_options) > 2:
            do_nn_filter = True
            model_path, overlap, nn_batch_size, filter_img_path = filter_options
            if not otf:
                print("Filter data using noise2noise model: ", model_path)
                img_paths = utils.filter_images_nn_dir(
                    img_paths=img_paths,
                    output_dir_filtered_imgs=filter_img_path,
                    model_path=model_path,
                    padding=overlap,
                    batch_size=nn_batch_size,
                )
        else:
            # Normal lowpass filter
            cutoff, filter_img_path = filter_options
            if not otf:
                img_paths = filter_images_lowpass(
                    img_paths, filter_img_path, cutoff, num_cpus=num_cpus
                )

    ###############################
    #   Predict bounding boxes
    ###############################
    print("Reset progress bar: ( Progress: -1 % )")
    total_picked = 0
    boxes_to_write = []
    measured_times = []
    picked_img = 1
    batchsize = num_patches * num_patches * num_images_batch_prediction
    tiles = []
    img_tiles = None
    num_written_tiles = 0
    image_indices = []
    skipped_images = []
    for current_index_image, img_pth in enumerate(img_paths):

        if os.path.basename(img_pth)[0] != ".":

            start = time.time()

            # Read image file!
            try:

                if otf and "filter" in config["model"]:
                    if do_nn_filter:
                        print("Filter", img_pth)
                        image = utils.filter_images_nn_img(
                            img_pth,
                            model_path,
                            padding=overlap,
                            batch_size=nn_batch_size,
                        )
                    else:
                        image = filter_images_lowpass(
                            [img_pth], None, cutoff, num_cpus=num_cpus, otf=True
                        )[0]
                else:
                    image = imagereader.image_read(img_pth)
            except ValueError:
                print("Image not valid: ", img_pth, "SKIPPED")
                skipped_images.append(img_pth)
                continue

            if image is not None or image.shape[0] == 0 or image.shape[1] == 0:
                image_indices.append(current_index_image)

                for patch_x in np.arange(0, num_patches):
                    for patch_y in np.arange(0, num_patches):
                        tile_coordinates = imagereader.get_tile_coordinates(
                            image.shape[1],
                            image.shape[0],
                            num_patches,
                            (patch_x, patch_y),
                            overlap=overlap,
                        )
                        tiles.append(tile_coordinates)
                        img_tmp = image[tile_coordinates[1], tile_coordinates[0]]

                        if img_tiles is None:
                            number_tiles_left = (
                                (len(img_paths) - current_index_image)
                                * num_patches
                                * num_patches
                            )
                            num_tiles = min(batchsize, number_tiles_left)
                            img_tiles = np.empty(
                                shape=(num_tiles, img_tmp.shape[0], img_tmp.shape[1]),
                                dtype=np.float32,
                            )

                        img_tiles[num_written_tiles, :, :] = img_tmp
                        num_written_tiles = num_written_tiles + 1

                if num_written_tiles == batchsize or current_index_image == (
                    len(img_paths) - 1
                ):
                    if filament_mode:
                        nms_thresh = 0.5
                    else:
                        nms_thresh = 0.3
                    boxes_per_image_nms, boxes_per_image_unfiltered = yolo.predict(
                        img_tiles,
                        tiles,
                        image.shape,
                        obj_threshold=obj_threshold,
                        nms_threshold=nms_thresh,
                        num_patches=num_patches,
                        normalize_margin=normalization_margin,
                    )
                    end = time.time()
                    measured_times.append(end - start)
                    for box_img_index, boxes in enumerate(boxes_per_image_nms):
                        boxes_image_path = img_paths[image_indices[box_img_index]]
                        imgh, imgw = imagereader.read_width_height(boxes_image_path)
                        total = len(img_paths)
                        print(
                            len(boxes),
                            "particles are found in",
                            boxes_image_path,
                            " ( Progress: ",
                            int(float(picked_img) * 100 / len(img_paths)),
                            "% )",
                        )

                        picked_img += 1
                        total_picked = total_picked + len(boxes)

                        if boxes or write_empty:
                            box_pth = os.path.join(
                                os.path.dirname(boxes_image_path), "box"
                            )
                            file_name_without_extension = os.path.splitext(
                                os.path.basename(boxes_image_path)
                            )[0]
                            box_pth = os.path.join(
                                box_pth,
                                os.path.basename(file_name_without_extension) + ".box",
                            )

                            box_to_write = {
                                "pth": box_pth,
                                "img_width": imgw,
                                "img_height": imgh,
                                "img_path": boxes_image_path,
                                "boxes": boxes,
                                "boxes_unfiltered": boxes_per_image_unfiltered[
                                    box_img_index
                                ],
                            }
                            boxes_to_write.append(box_to_write)
                        else:
                            print("no boxes: ", boxes_image_path)

                    # Reset variables
                    tiles = []
                    img_tiles = None
                    num_written_tiles = 0
                    image_indices = []
        else:
            print("Not a valid image:", img_pth)
    if len(skipped_images) > 0:
        print(
            "The following images were skipped because of errors during reading them:"
        )
        for img in skipped_images:
            print(img)
    print(
        total_picked,
        "particles in total are found",
        "(",
        int(np.sum(measured_times)),
        "seconds)",
    )

    return boxes_to_write


def get_not_fully_immersed_box_indices(boxes, image_height, image_width):
    boxes_to_delete = []
    for box_index, box in enumerate(boxes):
        if (
            (box.x + box.w) >= image_height
            or box.x < 0
            or box.y < 0
            or (box.y + box.h) >= image_width
        ):
            boxes_to_delete.append(box_index)
    return boxes_to_delete


def filter_images_lowpass(
    img_paths, output_dir_filtered_imgs, cutoff, num_cpus=-1, otf=False
):
    """
    Filteres a list of images and return a new list with the paths to the filtered images.

    :param img_paths: Path to images to filter
    :param output_dir_filtered_imgs: Output directory to save the filtered images
    :param cutoff: Absolute cutoff frequency (0-0.5)
    :return: List of paths to the filtered images
    """
    if not otf and not os.path.isdir(output_dir_filtered_imgs):
        os.makedirs(output_dir_filtered_imgs)

    arg_tubles = []
    for img_id, img_pth in enumerate(img_paths):
        if os.path.basename(img_pth)[0] != ".":
            if otf:
                arg_tubles.append((img_pth, cutoff))
            else:
                arg_tubles.append(
                    (
                        img_pth,
                        cutoff,
                        output_dir_filtered_imgs,
                        img_id + 1,
                        len(img_paths),
                    )
                )

    num_processes = None
    if num_cpus != -1:
        num_processes = num_cpus
    lock = multiprocessing.Lock()
    pool = multiprocessing.Pool(
        initializer=lowpass.init,
        initargs=(lock,),
        maxtasksperchild=1,
        processes=num_processes,
    )

    if otf:
        filtered_images = pool.starmap(
            lowpass.filter_single_image, arg_tubles, chunksize=1
        )
    else:
        filtered_images = pool.starmap(
            lowpass.filter_single_image_and_write_to_disk, arg_tubles, chunksize=1
        )
    pool.close()
    pool.join()

    return filtered_images


def trace_subset_filements(tracer_index):
    return filament_tracers[tracer_index].trace_filaments()


if __name__ == "__main__":
    _main_()



#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

from keras.callbacks import ModelCheckpoint
import h5py


class MyModelCheckpoint(ModelCheckpoint):

    def __init__(self, filepath, monitor='val_loss', verbose=0,
             save_best_only=False, save_weights_only=False,
             mode='auto', period=1):
        pass

    def on_epoch_end(self, epoch, logs=None):
        current = logs.get(self.monitor)
        model_improved = self.monitor_op(current, self.best)
        super(ModelCheckpoint, self).on_epoch_end(epoch, logs)
        if model_improved:

            with h5py.File(self.filepath, mode='r+') as f:
                f["anchors"] = self.anchors





from gooey import Gooey, GooeyParser
import argparse
import os
import cryolo.config_tools as config_tools
import multiprocessing
import cryolo.__init__ as ini

def config_configuration_parser(parser_config):
    config_required_group = parser_config.add_argument_group(
        "Required arguments",
        "The arguments are required to create a config file for crYOLO",
    )

    config_required_group.add_argument(
        "config_out_path",
        default="config_cryolo.json",
        help="Path where you want to write the config file.",
        widget="FileSaver",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("json")',
                "message": "File has to end with .json!",
            },
            "default_file":"config_cryolo.json"
        },
    )
    config_required_group.add_argument(
        "boxsize",
        type=int,
        help="You should specifiy your box size here. If you train on several datasets, use something like the average.",
    )

    config_required_group.add_argument(
        "train_image_folder",
        help="Path to the image folder containing the images to train on.",
        widget="DirChooser",
    )
    config_required_group.add_argument(
        "train_annot_folder",
        help="Path to folder containing the your annotation files like box or star files.",
        widget="DirChooser",
    )

    config_required_group.add_argument(
        "--saved_weights_name",
        default="cryolo_model.h5",
        help="Path for saving final weights.",
        widget="FileSaver",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            },
            "default_file":"cryolo_model.h5"
        },
    )

    config_required_group.add_argument(
        "--pretrained_weights",
        default="",
        help="Path to h5 file that is used for initialization in case you want to fine tune a previous model.",
        widget="FileChooser",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            }
        },
    )

    config_model_group = parser_config.add_argument_group(
        "Model/Denoising options", "Model configuration and denoising options"
    )

    config_model_group.add_argument(
        "-a",
        "--architecture",
        default="PhosaurusNet",
        choices=["PhosaurusNet", "YOLO", "crYOLO"],
        help="Backend network architecture.",
    )
    config_model_group.add_argument(
        "--input_size",
        default=1024,
        type=int,
        help="The images are downsized to this size.",
        gooey_options={
            "validator": {
                "test": "float(user_input)%32 < 1",
                "message": "Input size has to be multiple of 32!",
            }
        },
    )
    config_model_group.add_argument(
        "-f",
        "--filter",
        default="LOWPASS",
        help="Noise filter applied before training/picking. You can choose between a normal lowpass filter"
        " and neurnal network denoising (JANNI).",
        choices=["NONE", "LOWPASS", "JANNI"],
    )

    config_model_group.add_argument(
        "--low_pass_cutoff",
        type=float,
        default=0.1,
        help="Low pass filter cutoff frequency",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 0.5",
                "message": "Must be between 0 and 0.5",
            }
        },
    )
    config_model_group.add_argument(
        "--janni_model",
        help="Path to JANNI model",
        widget="FileChooser",
        default=None
    )

    config_model_group.add_argument(
        "--janni_overlap",
        type=int,
        default=24,
        help="Overlap of patches in pixel (only needed when using JANNI)",
    )

    config_model_group.add_argument(
        "--janni_batches",
        type=int,
        default=3,
        help="Number of batches (only needed when using JANNI)",
    )

    config_model_group.add_argument(
        "--filtered_output",
        default="filtered_tmp/",
        help="Output folder for filtered images",
        widget="DirChooser",
    )
    config_model_group.add_argument(
        "--num_patches",
        type=int,
        default=1,
        help="If specified the patch mode will be used. A value of “2” means, that 2×2 patches will be used.",
    )
    config_model_group.add_argument(
        "--overlap_patches",
        type=int,
        default=200,
        help="Only needed when using patch mode. Specifies how much the patches overlap. In our lab, we always keep the default value.",
    )

    config_training_group = parser_config.add_argument_group(
        "Training constants", "Training configuration"
    )

    config_training_group.add_argument(
        "--train_times",
        default=10,
        type=int,
        help="How often each image is presented to the network during one epoch. The default should be kept until you have many training images.",
    )

    config_training_group.add_argument(
        "--batch_size",
        type=int,
        default=4,
        help="The number of images crYOLO process in parallel during training.",
    )
    config_training_group.add_argument(
        "--learning_rate",
        type=float,
        default=10 ** -4,
        help="Defines the step size during training. Default should be kept.",
    )
    config_training_group.add_argument(
        "--nb_epoch",
        type=int,
        default=200,
        help="Maximum number of epochs the network will train.",
    )
    config_training_group.add_argument(
        "--object_scale",
        type=float,
        default=5.0,
        help="Penality scaling factor for missing picking particles.",
    )
    config_training_group.add_argument(
        "--no_object_scale",
        type=float,
        default=1.0,
        help="Penality scaling factor for picking background.",
    )
    config_training_group.add_argument(
        "--coord_scale",
        type=float,
        default=1.0,
        help="Penality scaling factor for errors in estimating the correct position.",
    )
    config_training_group.add_argument(
        "--class_scale",
        type=float,
        default=1.0,
        help='Irrelevant, as crYOLO only has the class "particle".',
    )
    config_training_group.add_argument(
        "--log_path", default="logs/", help="Path for log saving", widget="DirChooser"
    )

    config_training_group.add_argument(
        "--debug",
        action="store_true",
        default=True,
        help="If true, the network will provide several statistics during training.",
    )

    config_validation_group = parser_config.add_argument_group(
        "Validation configuration",
        "Validation configuration. If not specified, crYOLO will simply select 20% of the training data for validation. However it is possible to specify to use specific images for validation.",
    )
    config_validation_group.add_argument(
        "--valid_image_folder",
        default="",
        help="Path to folder containing the image files",
        widget="DirChooser",
    )
    config_validation_group.add_argument(
        "--valid_annot_folder",
        default="",
        help="Path to folder containing the validation box files",
        widget="DirChooser",
    )


def create_parser():
    parent_parser = GooeyParser(description="Let crYOLO pick your particles!")
    subparsers = parent_parser.add_subparsers(help="sub-command help")

    # Config generator
    parser_config = subparsers.add_parser("config", help="train help")
    config_configuration_parser(parser_config)

    # Training parser
    parser_train = subparsers.add_parser("train", help="train help")
    import cryolo.train as train

    train.create_parser(parser_train)

    # Picking parser
    parser_pick = subparsers.add_parser("predict", help="predict help")
    import cryolo.predict as predict

    predict.create_parser(parser_pick)

    # Evaluation parser
    parser_eval = subparsers.add_parser("evaluation", help="evalutation help")
    import cryolo.eval as evaluation

    evaluation.create_parser(parser_eval)

    return parent_parser


PARSER = None


def main():
    import sys

    args = PARSER.parse_args()
    if "config" in sys.argv[1]:
        filter = None
        if args.filter == "LOWPASS":
            filter = [args.low_pass_cutoff, args.filtered_output]
        elif args.filter == "JANNI":
            if args.janni_model is None:
                print("Please specify the JANNI model file.")
                sys.exit(1)
            filter = [args.janni_model, args.janni_overlap, args.janni_batches, args.filtered_output]

        config_tools.generate_config_file(
            config_out_path=args.config_out_path,
            architecture=args.architecture,
            input_size=args.input_size,
            anchors=[args.boxsize, args.boxsize],
            max_box_per_image=700,
            num_patches=args.num_patches,
            overlap_patches=args.overlap_patches,
            filter=filter,
            train_image_folder=args.train_image_folder,
            train_annot_folder=args.train_annot_folder,
            train_times=args.train_times,
            pretrained_weights=args.pretrained_weights,
            batch_size=args.batch_size,
            learning_rate=args.learning_rate,
            nb_epoch=args.nb_epoch,
            object_scale=args.object_scale,
            no_object_scale=args.no_object_scale,
            coord_scale=args.coord_scale,
            class_scale=args.class_scale,
            log_path=args.log_path,
            saved_weights_name=args.saved_weights_name,
            debug=args.debug,
            valid_image_folder=args.valid_image_folder,
            valid_annot_folder=args.valid_annot_folder,
            valid_times=1,
        )
    elif "train" in sys.argv[1]:
        import cryolo.train as train
        train.main(args)
    elif "predict" in sys.argv[1]:
        import cryolo.predict as predict
        predict.main(args)
    elif "evaluation" in sys.argv[1]:
        import cryolo.eval as evaluation
        evaluation.main(args)


def _main_():
    global PARSER
    PARSER = create_parser()
    try:
        multiprocessing.set_start_method("spawn")
    except RuntimeError:
        print("Ignore set start method")

    kwargs = {"terminal_font_family":"monospace",
              "richtext_controls": True}



    Gooey(
        main,
        program_name="crYOLO " + ini.__version__,
        image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        default_size=(1024, 530),
        **kwargs
    )()


if __name__ == "__main__":
    try:
        multiprocessing.set_start_method("spawn")
    except RuntimeError:
        print("Ignore set start method")
    _main_()



from setuptools import setup
import os
import re
import codecs
import sys
import setuptools
from distutils.command.sdist import sdist

# Create new package with python setup.py sdist



here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    print("==========================================",sys.argv)
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cryolo',
    version=find_version("cryolo", "__init__.py"),
    python_requires='>3.4.0',
    packages=['cryolo'],
    url='http://sphire.mpg.de/wiki/doku.php?id=downloads:cryolo_1',
    license='Other/Proprietary License (all rights reserved)',
    author='Thorsten Wagner',
    include_package_data=True,
    package_data={'cryolo': ['../icons/config_icon.png','../icons/program_icon.ico']},
    setup_requires=["Cython"],
    extras_require = {
        'gpu':  ['tensorflow-gpu == 1.10.1','janni[gpu] >= 0.0.5'],
        'cpu': ['tensorflow == 1.10.1', 'janni[cpu] >= 0.0.5']
    },
    install_requires=[
        "mrcfile >= 1.0.0",
        "Cython",
        "Keras >= 2.2.4",
        "numpy == 1.14.5",
        "h5py >= 2.5.0",
        "imageio >= 2.3.0",
        "Pillow >= 6.0.0",
        "tifffile",
        "scipy >= 1.3.0",
        "terminaltables",
        "lineenhancer >= 1.0.6",
        "cryoloBM >= 1.2.6",
        "ansi2html",
        "GooeyDev >= 1.0.3.1"
    ],
    author_email='thorsten.wagner@mpi-dortmund.mpg.de',
    description='Picking procedure for cryo em single particle analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'cryolo_train.py = cryolo.train:_main_',
            'cryolo_predict.py = cryolo.predict:_main_',
            'cryolo_evaluation.py = cryolo.eval:_main_',
            'cryoloo.py = cryolo.cryolo_main:_main_']},
)
